/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "vfat_iter.h"
#include "vfat_iter_struct.h"
#include "vfat_attr_struct.h"
#include "vfat_names.h"
#include "pfs.h"

#include <segment.h>
#include <consts.h>
#include <errno.h>
#include <config.h>
#include <string.h>
#include <util.h>

/* Utilitaires (time) */

static inline int datecmp(const struct calendar_time_struct *calendar_time1, const struct calendar_time_struct *calendar_time2)
{
  if (calendar_time1->year != calendar_time2->year)
    return calendar_time1->year < calendar_time2->year ? -1 : 1;
  if (calendar_time1->mon != calendar_time2->mon)
    return calendar_time1->mon < calendar_time2->mon ? -1 : 1;
  if (calendar_time1->mday != calendar_time2->mday)
    return calendar_time1->mday < calendar_time2->mday ? -1 : 1;
  return 0;
}

static inline int timecmp(const struct calendar_time_struct *calendar_time1, const struct calendar_time_struct *calendar_time2)
{
  register int tmp;

  if ((tmp = datecmp(calendar_time1, calendar_time2)))
    return tmp;
  if (calendar_time1->hour != calendar_time2->hour)
    return calendar_time1->hour < calendar_time2->hour ? -1 : 1;
  if (calendar_time1->min != calendar_time2->min)
    return calendar_time1->min < calendar_time2->min ? -1 : 1;
  if (calendar_time1->sec != calendar_time2->sec)
    return calendar_time1->sec < calendar_time2->sec ? -1 : 1;
  return 0;
}

static inline void make_calendar_time(unsigned short date, unsigned short time, unsigned short time_tenth, struct calendar_time_struct *calendar_time)
{
  calendar_time->sec = ((time & 0x1f) << 1) + time_tenth / 100;
  calendar_time->min = (time & 0x7e0) >> 5;
  calendar_time->hour = (time & 0xf800) >> 11;
  calendar_time->mday = date & 0x1f;
  calendar_time->mon = (date & 0x1e0) >> 5;
  calendar_time->year = 1980 + ((date & 0xfe00) >> 9);
}

static inline unsigned short make_fat_time_tenth(const struct calendar_time_struct *calendar_time)
{
  return calendar_time->sec & 1 ? 100 : 0;
}

static inline unsigned short make_fat_time(const struct calendar_time_struct *calendar_time)
{
  return (calendar_time->hour << 11) | (calendar_time->min << 5) | (calendar_time->sec >> 1);
}

static inline unsigned short make_fat_date(const struct calendar_time_struct *calendar_time)
{
  return ((calendar_time->year - 1980) << 9) | (calendar_time->mon << 5) | calendar_time->mday;
}

/* Utilitaires (noms) */

static inline unsigned int wstrlen(const unsigned short *str)
{
  unsigned int len;

  len = 0;
  while (*str++)
    len++;
  return len;
}

static inline int match(const char *fs_basename, const unsigned short longname[], const unsigned char shortname[])
{
  return (longname && vfat_match_long(fs_basename, longname))
    || (vfat_match_short(fs_basename, shortname));
}

static inline void decode(const unsigned short longname[], const unsigned char shortname[], char name[])
{
  if (longname)
    vfat_decode_longname(longname, name);
  else
    vfat_decode_shortname(shortname, name);
}

/* Utilitaires (entrees de repertoire) */

static inline int is_dot(const directoryEntryStructure_t *dir_ent)
{
  return (dir_ent->DIR_Attr & (ATTR_DIRECTORY | ATTR_VOLUME_ID)) == ATTR_DIRECTORY && !memcmp(dir_ent->DIR_Name, ".          ", 11);
}

static inline int is_dotdot(const directoryEntryStructure_t *dir_ent)
{
  return (dir_ent->DIR_Attr & (ATTR_DIRECTORY | ATTR_VOLUME_ID)) == ATTR_DIRECTORY && !memcmp(dir_ent->DIR_Name, "..         ", 11);
}

/* Dit si une entree courte de repertoire est celle d'un fichier (regulier ou
 * repertoire). */
static inline int is_file_dir_entry(const directoryEntryStructure_t *dir_ent)
{
  if (is_dot(dir_ent) || is_dotdot(dir_ent))
    return 0;
  switch (dir_ent->DIR_Attr & (ATTR_DIRECTORY | ATTR_VOLUME_ID)) {
  case 0x00:           /* fichier regulier */
  case ATTR_DIRECTORY: /* repertoire */
    return 1;
  case ATTR_VOLUME_ID: /* libelle de volume */
    return 0;
  default:             /* entree invalide */
    return 0;
  }
}

/* La specification specifie qu'un nom long est limite a 255 caracteres (sans
 * compter le zero terminal) et que le chemin total d'un nom long est limite a
 * 260 caracteres (en comptant le zero terminal). Dans notre implementation, le
 * nom long est tronque s'il fait plus de 255 caracteres, et nous ignorons la
 * seconde limitation. */
static void read_long_name(const longDirectoryEntryStructure_t *dir_ent, unsigned short longname[])
{
  unsigned char ord;
  int i, j;

  ord = dir_ent->LDIR_Ord & ~LAST_LONG_ENTRY;
  if (dir_ent->LDIR_Ord & LAST_LONG_ENTRY) { /* la "derniere" */
    j = (ord - 1) * 13;
    for (i = 0; i < 5; i++) {
      if (!dir_ent->LDIR_Name1[i] || j >= 255)
        goto end_longname;
      longname[j++] = dir_ent->LDIR_Name1[i];
    }
    for (i = 0; i < 6; i++) {
      if (!dir_ent->LDIR_Name2[i] || j >= 255)
        goto end_longname;
      longname[j++] = dir_ent->LDIR_Name2[i];
    }
    for (i = 0; i < 2; i++) {
      if (!dir_ent->LDIR_Name3[i] || j >= 255)
        goto end_longname;
      longname[j++] = dir_ent->LDIR_Name3[i];
    }
  end_longname:
    longname[j] = '\0';
  }
  else {
    memcpy(longname + (ord - 1) * 13, dir_ent->LDIR_Name1, 10);
    memcpy(longname + (ord - 1) * 13 + 5, dir_ent->LDIR_Name2, 12);
    memcpy(longname + (ord - 1) * 13 + 11, dir_ent->LDIR_Name3, 4);
  }
}

static void write_long_name(longDirectoryEntryStructure_t *dir_ent, const unsigned short longname[])
{
  unsigned char ord;
  int pad;
  int i, j;

  ord = dir_ent->LDIR_Ord & ~LAST_LONG_ENTRY;
  if (dir_ent->LDIR_Ord & LAST_LONG_ENTRY) { /* la "derniere" */
    pad = 0;
    j = (ord - 1) * 13;
    for (i = 0; i < 5; i++) {
      if (pad)
        dir_ent->LDIR_Name1[i] = 0xffff;
      else if (j > 255 || !longname[j]) {
        dir_ent->LDIR_Name1[i] = '\0';
        pad = 1;
      }
      else
        dir_ent->LDIR_Name1[i] = longname[j++];
    }
    for (i = 0; i < 6; i++) {
      if (pad)
        dir_ent->LDIR_Name2[i] = 0xffff;
      else if (j > 255 || !longname[j]) {
        dir_ent->LDIR_Name2[i] = '\0';
        pad = 1;
      }
      else
        dir_ent->LDIR_Name2[i] = longname[j++];
    }
    for (i = 0; i < 2; i++) {
      if (pad)
        dir_ent->LDIR_Name3[i] = 0xffff;
      else if (j > 255 || !longname[j]) {
        dir_ent->LDIR_Name3[i] = '\0';
        pad = 1;
      }
      else
        dir_ent->LDIR_Name3[i] = longname[j++];
    }
  }
  else {
    memcpy(dir_ent->LDIR_Name1, longname + (ord - 1) * 13, 10);
    memcpy(dir_ent->LDIR_Name2, longname + (ord - 1) * 13 + 5, 12);
    memcpy(dir_ent->LDIR_Name3, longname + (ord - 1) * 13 + 11, 4);
  }
}

struct longname_reader_struct {
  int has_long_name;
  unsigned char ord;
  unsigned char chk_sum;
  unsigned short longname[256];
};

static void longname_reader_init(struct longname_reader_struct *longname_reader)
{
  longname_reader->has_long_name = 0;
}

static void longname_reader_do_free_entry(struct longname_reader_struct *longname_reader)
{
  longname_reader->has_long_name = 0;
}

static void longname_reader_do_long_dir_entry(struct longname_reader_struct *longname_reader, const longDirectoryEntryStructure_t *dir_ent)
{
  if (dir_ent->LDIR_Ord & LAST_LONG_ENTRY) { /* la "derniere" */
    longname_reader->has_long_name = 1;
    longname_reader->ord = dir_ent->LDIR_Ord & ~LAST_LONG_ENTRY;
    longname_reader->chk_sum = dir_ent->LDIR_Chksum;
  }
  else {
    if (!longname_reader->has_long_name)
      return;
    if (dir_ent->LDIR_Ord != longname_reader->ord - 1)
      goto invalid_long_entry;
    if (dir_ent->LDIR_Chksum != longname_reader->chk_sum)
      goto invalid_long_entry;
    longname_reader->ord--;
  }
  if (longname_reader->ord == 0)
    goto invalid_long_entry;
  if (dir_ent->LDIR_Type)
    goto invalid_long_entry;
  read_long_name(dir_ent, longname_reader->longname);
  return;
 invalid_long_entry: /* orphan */
  longname_reader->has_long_name = 0;
}

/* La validite du nom long est verifiee. Si celui-ci est invalide, ou n'existe
 * pas, NULL est retourne. */
static unsigned short *longname_reader_do_short_dir_entry(struct longname_reader_struct *longname_reader, const directoryEntryStructure_t *dir_ent)
{
  if (longname_reader->has_long_name) {
    longname_reader->has_long_name = 0;
    if (longname_reader->ord != 1)
      goto no_long_name;
    if (longname_reader->chk_sum != chkSum(dir_ent->DIR_Name))
      goto no_long_name;
    return longname_reader->longname;
  }
  else {
  no_long_name:
    return NULL;
  }
}

struct longname_writer_struct {
  const unsigned short *longname; /* nul si pas de nom long */
  unsigned char ord;
  int last;
  unsigned char chk_sum;
};

/* longname doit faire au plus 255 caracteres. */
static void longname_writer_init(struct longname_writer_struct *longname_writer, const unsigned short longname[], unsigned char chk_sum)
{
  longname_writer->longname = longname;
  if (longname) {
    longname_writer->ord = (wstrlen(longname) + 12) / 13;
    longname_writer->last = 1;
    longname_writer->chk_sum = chk_sum;
  }
  else
    longname_writer->ord = 0;
}

static int longname_writer_do_long_dir_entry(struct longname_writer_struct *longname_writer, longDirectoryEntryStructure_t *dir_ent)
{
  if (longname_writer->ord) {
    dir_ent->LDIR_Attr = ATTR_LONG_NAME;
    dir_ent->LDIR_Type = 0;
    dir_ent->LDIR_Chksum = longname_writer->chk_sum;
    dir_ent->LDIR_FstClusLO = 0;
    if (longname_writer->last) {
      dir_ent->LDIR_Ord = LAST_LONG_ENTRY | longname_writer->ord;
      longname_writer->last = 0;
    }
    else
      dir_ent->LDIR_Ord = longname_writer->ord;
    write_long_name(dir_ent, longname_writer->longname);
    longname_writer->ord--;
    return 1;
  }
  else
    return 0;
}

/* traverse_dir */

struct traverse_dir_fo_struct;

struct traverse_dir_fo_vtbl_struct {
  int (*do_dir_entry)(struct traverse_dir_fo_struct *traverse_dir_fo, const struct vfat_dir_iter_struct *dir_iter);
  int (*get_retval)(struct traverse_dir_fo_struct *traverse_dir_fo);
};

struct traverse_dir_fo_struct {
  const struct traverse_dir_fo_vtbl_struct *vptr;
};

static int traverse_dir(struct traverse_dir_fo_struct *traverse_dir_fo, struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key)
{
  int retval;
  struct vfat_dir_iter_struct dir_it;

  if ((retval = vfat_init_dir_iter(&dir_it, vfat_instance, dir_ent_key)) < 0)
    goto end;
  do {
    if (!traverse_dir_fo->vptr->do_dir_entry(traverse_dir_fo, &dir_it))
      break;
    if ((retval = vfat_next_dir_iter(&dir_it)) < 0)
      goto destroy;
  }
  while (retval);
  retval = traverse_dir_fo->vptr->get_retval(traverse_dir_fo);
 destroy:
  vfat_destroy_dir_iter(&dir_it);
 end:
  return retval;
}

/* traverse_file */

struct traverse_file_fo_struct;

struct traverse_file_fo_vtbl_struct {
  int (*do_file_blk)(struct traverse_file_fo_struct *traverse_file_fo, const struct vfat_file_iter_struct *file_iter);
  int (*get_retval)(struct traverse_file_fo_struct *traverse_file_fo);
};

struct traverse_file_fo_struct {
  const struct traverse_file_fo_vtbl_struct *vptr;
};

static int traverse_file(struct traverse_file_fo_struct *traverse_file_fo, struct vfat_instance_struct *vfat_instance, unsigned long file_blk_key)
{
  int retval;
  struct vfat_file_iter_struct file_it;

  if ((retval = vfat_init_file_iter(&file_it, vfat_instance, file_blk_key)) < 0)
    goto end;
  do {
    if (!traverse_file_fo->vptr->do_file_blk(traverse_file_fo, &file_it))
      break;
    if ((retval = vfat_next_file_iter(&file_it)) < 0)
      goto destroy;
  }
  while (retval);
  retval = traverse_file_fo->vptr->get_retval(traverse_file_fo);
 destroy:
  vfat_destroy_file_iter(&file_it);
 end:
  return retval;
}

/* Fonctions-objets pour traverse_dir */

/* traverse_entries (abstraite) */
/* Appelle une fonction de rappel differente suivant le type d'entree. */

struct traverse_entries_fo_struct;

struct traverse_entries_fo_vtbl_struct {
  int (*do_free_entry)(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter);
  int (*do_free_entry_no_more)(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter);
  int (*do_long_dir_entry)(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter);
  int (*do_short_dir_entry)(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter);
};

struct traverse_entries_fo_struct {
  struct traverse_dir_fo_struct traverse_dir_fo;
  const struct traverse_entries_fo_vtbl_struct *vptr;
};

static int traverse_entries_do_dir_entry(struct traverse_dir_fo_struct *traverse_dir_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct traverse_entries_fo_struct *traverse_entries_fo;
  union vfat_dir_entry_union *dir_ent;

  traverse_entries_fo = downcast(struct traverse_entries_fo_struct, traverse_dir_fo, traverse_dir_fo);
  dir_ent = vfat_get_dir_iter_dir_entry(dir_iter);
  if (dir_ent->short_struct.DIR_Name[0] == 0xe5)      /* libre */
    return traverse_entries_fo->vptr->do_free_entry(traverse_entries_fo, dir_iter);
  else if (dir_ent->short_struct.DIR_Name[0] == 0x00) /* libre, plus d'entree allouee apres */
    return traverse_entries_fo->vptr->do_free_entry_no_more(traverse_entries_fo, dir_iter);
#if VFAT_USE_LONG_NAMES != 0
  else if ((dir_ent->short_struct.DIR_Attr & ATTR_LONG_NAME_MASK) == ATTR_LONG_NAME) /* long directory entry */
    return traverse_entries_fo->vptr->do_long_dir_entry(traverse_entries_fo, dir_iter);
#endif
  else                                                /* short directory entry */
    return traverse_entries_fo->vptr->do_short_dir_entry(traverse_entries_fo, dir_iter);
}

static int traverse_entries_skip_free_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  return 1;
}

static int traverse_entries_stop_free_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  return 0;
}

static int traverse_entries_stop_free_entry_no_more(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  return 0;
}

static int traverse_entries_skip_long_dir_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  return 1;
}

/* traverse_names (abstraite) */
/* Parcours de repertoire.
 * Le parcours s'arrete lorsqu'il n'y a plus d'entree allouee. */

struct traverse_names_fo_struct;

struct traverse_names_fo_vtbl_struct {
  void (*do_free_entry)(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter);
  void (*do_long_dir_entry)(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter);
  int (*do_short_dir_entry)(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter);
  int (*get_retval)(struct traverse_names_fo_struct *traverse_names_fo, int finished);
};

struct traverse_names_fo_struct {
  struct traverse_entries_fo_struct traverse_entries_fo;
  const struct traverse_names_fo_vtbl_struct *vptr;
  int state; /* 0=parcours en cours, 1=parcours interrompu, 2=parcours termine */
};

static int traverse_names_get_retval(struct traverse_dir_fo_struct *traverse_dir_fo)
{
  struct traverse_entries_fo_struct *traverse_entries_fo;
  struct traverse_names_fo_struct *traverse_names_fo;

  traverse_entries_fo = downcast(struct traverse_entries_fo_struct, traverse_dir_fo, traverse_dir_fo);
  traverse_names_fo = downcast(struct traverse_names_fo_struct, traverse_entries_fo, traverse_entries_fo);
  switch (traverse_names_fo->state) {
  case 0: /* parcours en cours */
    return -EINVAL;
  case 1: /* parcours interrompu */
    return traverse_names_fo->vptr->get_retval(traverse_names_fo, 0);
  case 2: /* parcours termine */
  default:
    return traverse_names_fo->vptr->get_retval(traverse_names_fo, 1);
  }
}

static int traverse_names_do_free_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct traverse_names_fo_struct *traverse_names_fo;

  traverse_names_fo = downcast(struct traverse_names_fo_struct, traverse_entries_fo, traverse_entries_fo);
  traverse_names_fo->vptr->do_free_entry(traverse_names_fo, dir_iter);
  return 1;
}

static int traverse_names_do_free_entry_no_more(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct traverse_names_fo_struct *traverse_names_fo;

  traverse_names_fo = downcast(struct traverse_names_fo_struct, traverse_entries_fo, traverse_entries_fo);
  traverse_names_fo->state = 2;
  return 0;
}

static int traverse_names_do_long_dir_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct traverse_names_fo_struct *traverse_names_fo;

  traverse_names_fo = downcast(struct traverse_names_fo_struct, traverse_entries_fo, traverse_entries_fo);
  traverse_names_fo->vptr->do_long_dir_entry(traverse_names_fo, dir_iter);
  return 1;
}

static int traverse_names_do_short_dir_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct traverse_names_fo_struct *traverse_names_fo;
  directoryEntryStructure_t *dir_ent;

  traverse_names_fo = downcast(struct traverse_names_fo_struct, traverse_entries_fo, traverse_entries_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  if (!is_file_dir_entry(dir_ent))
    return 1; /* ignore */
  if (!traverse_names_fo->vptr->do_short_dir_entry(traverse_names_fo, dir_iter)) {
    traverse_names_fo->state = 1;
    return 0;
  }
  return 1;
}

static const struct traverse_dir_fo_vtbl_struct traverse_names_traverse_dir_fo_vtbl = {
  .do_dir_entry = traverse_entries_do_dir_entry,
  .get_retval = traverse_names_get_retval
};

static const struct traverse_entries_fo_vtbl_struct traverse_names_traverse_entries_fo_vtbl = {
  .do_free_entry = traverse_names_do_free_entry,
  .do_free_entry_no_more = traverse_names_do_free_entry_no_more,
  .do_long_dir_entry = traverse_names_do_long_dir_entry,
  .do_short_dir_entry = traverse_names_do_short_dir_entry
};

static void traverse_names_fo_init(struct traverse_names_fo_struct *traverse_names_fo)
{
  traverse_names_fo->state = 0;
}

static void traverse_names_skip_free_entry(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter) {}

static void traverse_names_skip_long_dir_entry(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter) {}

/* traverse_name (abstraite) */
/* Parcours de nom.
 * Le parcours s'arrete a la premiere entree courte trouvee.
 * La validite du nom long n'est pas verifiee. */

struct traverse_name_fo_struct;

struct traverse_name_fo_vtbl_struct {
  void (*do_long_dir_entry)(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter);
  int (*do_short_dir_entry)(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter);
};

struct traverse_name_fo_struct {
  struct traverse_entries_fo_struct traverse_entries_fo;
  const struct traverse_name_fo_vtbl_struct *vptr;
  int retval;
};

static int traverse_name_get_retval(struct traverse_dir_fo_struct *traverse_dir_fo)
{
  struct traverse_entries_fo_struct *traverse_entries_fo;
  struct traverse_name_fo_struct *traverse_name_fo;

  traverse_entries_fo = downcast(struct traverse_entries_fo_struct, traverse_dir_fo, traverse_dir_fo);
  traverse_name_fo = downcast(struct traverse_name_fo_struct, traverse_entries_fo, traverse_entries_fo);
  return traverse_name_fo->retval;
}

static int traverse_name_do_long_dir_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct traverse_name_fo_struct *traverse_name_fo;

  traverse_name_fo = downcast(struct traverse_name_fo_struct, traverse_entries_fo, traverse_entries_fo);
  traverse_name_fo->vptr->do_long_dir_entry(traverse_name_fo, dir_iter);
  return 1;
}

static int traverse_name_do_short_dir_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct traverse_name_fo_struct *traverse_name_fo;
  directoryEntryStructure_t *dir_ent;

  traverse_name_fo = downcast(struct traverse_name_fo_struct, traverse_entries_fo, traverse_entries_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  if (is_file_dir_entry(dir_ent))
    traverse_name_fo->retval = traverse_name_fo->vptr->do_short_dir_entry(traverse_name_fo, dir_iter);
  return 0;
}

static const struct traverse_dir_fo_vtbl_struct traverse_name_traverse_dir_fo_vtbl = {
  .do_dir_entry = traverse_entries_do_dir_entry,
  .get_retval = traverse_name_get_retval
};

static const struct traverse_entries_fo_vtbl_struct traverse_name_traverse_entries_fo_vtbl = {
  .do_free_entry = traverse_entries_stop_free_entry,
  .do_free_entry_no_more = traverse_entries_stop_free_entry_no_more,
  .do_long_dir_entry = traverse_name_do_long_dir_entry,
  .do_short_dir_entry = traverse_name_do_short_dir_entry
};

static void traverse_name_fo_init(struct traverse_name_fo_struct *traverse_name_fo)
{
  traverse_name_fo->retval = -EINVAL;
}

static void traverse_name_skip_long_dir_entry(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter) {}

/* find */

struct find_fo_struct {
  struct traverse_names_fo_struct traverse_names_fo;
  const char *fs_basename;
  unsigned long *id;
  int *type;
  struct longname_reader_struct longname_reader;
};

static int find_get_retval(struct traverse_names_fo_struct *traverse_names_fo, int finished)
{
  return !finished;
}

static void find_do_free_entry(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct find_fo_struct *find_fo;

  find_fo = downcast(struct find_fo_struct, traverse_names_fo, traverse_names_fo);
  longname_reader_do_free_entry(&find_fo->longname_reader);
}

static void find_do_long_dir_entry(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct find_fo_struct *find_fo;
  const longDirectoryEntryStructure_t *dir_ent;

  find_fo = downcast(struct find_fo_struct, traverse_names_fo, traverse_names_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->long_struct;
  if (dir_ent->LDIR_Ord & LAST_LONG_ENTRY)
    *find_fo->id = vfat_get_dir_iter_dir_ent_key(dir_iter);
  longname_reader_do_long_dir_entry(&find_fo->longname_reader, dir_ent);
}

static int find_do_short_dir_entry(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct find_fo_struct *find_fo;
  const directoryEntryStructure_t *dir_ent;
  const unsigned short *longname;

  find_fo = downcast(struct find_fo_struct, traverse_names_fo, traverse_names_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  longname = longname_reader_do_short_dir_entry(&find_fo->longname_reader, dir_ent);
  if (!longname)
    *find_fo->id = vfat_get_dir_iter_dir_ent_key(dir_iter);
  if (match(find_fo->fs_basename, longname, dir_ent->DIR_Name)) {
    *find_fo->type = dir_ent->DIR_Attr & ATTR_DIRECTORY ? FT_DIR : FT_REG;
    return 0;
  }
  else
    return 1;
}

static const struct traverse_names_fo_vtbl_struct find_traverse_names_fo_vtbl = {
  .do_free_entry = find_do_free_entry,
  .do_long_dir_entry = find_do_long_dir_entry,
  .do_short_dir_entry = find_do_short_dir_entry,
  .get_retval = find_get_retval
};

static void find_fo_init(struct find_fo_struct *find_fo, const char *fs_basename, unsigned long *id, int *type)
{
  find_fo->traverse_names_fo.traverse_entries_fo.traverse_dir_fo.vptr = &traverse_names_traverse_dir_fo_vtbl;
  find_fo->traverse_names_fo.traverse_entries_fo.vptr = &traverse_names_traverse_entries_fo_vtbl;
  find_fo->traverse_names_fo.vptr = &find_traverse_names_fo_vtbl;
  traverse_names_fo_init(&find_fo->traverse_names_fo);
  find_fo->fs_basename = fs_basename;
  find_fo->id = id;
  find_fo->type = type;
  longname_reader_init(&find_fo->longname_reader);
}

/* read_dir */

struct read_dir_fo_struct {
  struct traverse_names_fo_struct traverse_names_fo;
  unsigned long pos;
  char *fs_buf;
  unsigned int count;
  struct longname_reader_struct longname_reader;
  unsigned long base_pos;
  unsigned int n;
};

static int read_dir_get_retval(struct traverse_names_fo_struct *traverse_names_fo, int finished)
{
  struct read_dir_fo_struct *read_dir_fo;

  read_dir_fo = downcast(struct read_dir_fo_struct, traverse_names_fo, traverse_names_fo);
  return read_dir_fo->n;
}

static void read_dir_do_free_entry(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct read_dir_fo_struct *read_dir_fo;

  read_dir_fo = downcast(struct read_dir_fo_struct, traverse_names_fo, traverse_names_fo);
  longname_reader_do_free_entry(&read_dir_fo->longname_reader);
}

static void read_dir_do_long_dir_entry(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct read_dir_fo_struct *read_dir_fo;
  longDirectoryEntryStructure_t *dir_ent;

  read_dir_fo = downcast(struct read_dir_fo_struct, traverse_names_fo, traverse_names_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->long_struct;
  longname_reader_do_long_dir_entry(&read_dir_fo->longname_reader, dir_ent);
}

static int read_dir_do_short_dir_entry(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct read_dir_fo_struct *read_dir_fo;
  const directoryEntryStructure_t *dir_ent;
  const unsigned short *longname;
  char name[256];
  unsigned int len;

  read_dir_fo = downcast(struct read_dir_fo_struct, traverse_names_fo, traverse_names_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  longname = longname_reader_do_short_dir_entry(&read_dir_fo->longname_reader, dir_ent);
  decode(longname, dir_ent->DIR_Name, name);

  len = strlen(name) + 1;
  if (read_dir_fo->base_pos + len <= read_dir_fo->pos) {
    read_dir_fo->base_pos += len;
    return 1;
  }
  read_dir_fo->n = pfs_read_name(name, read_dir_fo->pos - read_dir_fo->base_pos, read_dir_fo->fs_buf, read_dir_fo->count);
  return 0;
}

static const struct traverse_names_fo_vtbl_struct read_dir_traverse_names_fo_vtbl = {
  .do_free_entry = read_dir_do_free_entry,
  .do_long_dir_entry = read_dir_do_long_dir_entry,
  .do_short_dir_entry = read_dir_do_short_dir_entry,
  .get_retval = read_dir_get_retval
};

static void read_dir_fo_init(struct read_dir_fo_struct *read_dir_fo, unsigned long pos, char *fs_buf, unsigned int count)
{
  read_dir_fo->traverse_names_fo.traverse_entries_fo.traverse_dir_fo.vptr = &traverse_names_traverse_dir_fo_vtbl;
  read_dir_fo->traverse_names_fo.traverse_entries_fo.vptr = &traverse_names_traverse_entries_fo_vtbl;
  read_dir_fo->traverse_names_fo.vptr = &read_dir_traverse_names_fo_vtbl;
  traverse_names_fo_init(&read_dir_fo->traverse_names_fo);
  read_dir_fo->pos = pos;
  read_dir_fo->fs_buf = fs_buf;
  read_dir_fo->count = count;
  longname_reader_init(&read_dir_fo->longname_reader);
  read_dir_fo->base_pos = 0;
  read_dir_fo->n = 0;
}

/* collide */

struct collide_fo_struct {
  struct traverse_names_fo_struct traverse_names_fo;
  const unsigned char *shortname;
};

static int collide_get_retval(struct traverse_names_fo_struct *traverse_names_fo, int finished)
{
  return !finished;
}

static int collide_do_short_dir_entry(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct collide_fo_struct *collide_fo;
  const directoryEntryStructure_t *dir_ent;

  collide_fo = downcast(struct collide_fo_struct, traverse_names_fo, traverse_names_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  return memcmp(collide_fo->shortname, dir_ent->DIR_Name, 11);
}

static const struct traverse_names_fo_vtbl_struct collide_traverse_names_fo_vtbl = {
  .do_free_entry = traverse_names_skip_free_entry,
  .do_long_dir_entry = traverse_names_skip_long_dir_entry,
  .do_short_dir_entry = collide_do_short_dir_entry,
  .get_retval = collide_get_retval
};

static void collide_fo_init(struct collide_fo_struct *collide_fo, const unsigned char *shortname)
{
  collide_fo->traverse_names_fo.traverse_entries_fo.traverse_dir_fo.vptr = &traverse_names_traverse_dir_fo_vtbl;
  collide_fo->traverse_names_fo.traverse_entries_fo.vptr = &traverse_names_traverse_entries_fo_vtbl;
  collide_fo->traverse_names_fo.vptr = &collide_traverse_names_fo_vtbl;
  traverse_names_fo_init(&collide_fo->traverse_names_fo);
  collide_fo->shortname = shortname;
}

/* clear_dir */

struct clear_dir_fo_struct {
  struct traverse_dir_fo_struct traverse_dir_fo;
};

static int clear_dir_do_dir_entry(struct traverse_dir_fo_struct *traverse_dir_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  vfat_dirty_dir_iter_buffer(dir_iter);
  vfat_get_dir_iter_dir_entry(dir_iter)->short_struct.DIR_Name[0] = 0x00;
  return 1;
}

static int clear_dir_get_retval(struct traverse_dir_fo_struct *traverse_dir_fo)
{
  return 0;
}

static const struct traverse_dir_fo_vtbl_struct clear_dir_traverse_dir_fo_vtbl = {
  .do_dir_entry = clear_dir_do_dir_entry,
  .get_retval = clear_dir_get_retval
};

static void clear_dir_fo_init(struct clear_dir_fo_struct *clear_dir_fo)
{
  clear_dir_fo->traverse_dir_fo.vptr = &clear_dir_traverse_dir_fo_vtbl;
}

/* is_empty */

struct is_empty_fo_struct {
  struct traverse_names_fo_struct traverse_names_fo;
};

static int is_empty_get_retval(struct traverse_names_fo_struct *traverse_names_fo, int finished)
{
  return finished;
}

static int is_empty_do_short_dir_entry(struct traverse_names_fo_struct *traverse_names_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  return 0;
}

static const struct traverse_names_fo_vtbl_struct is_empty_traverse_names_fo_vtbl = {
  .do_free_entry = traverse_names_skip_free_entry,
  .do_long_dir_entry = traverse_names_skip_long_dir_entry,
  .do_short_dir_entry = is_empty_do_short_dir_entry,
  .get_retval = is_empty_get_retval
};

static void is_empty_fo_init(struct is_empty_fo_struct *is_empty_fo)
{
  is_empty_fo->traverse_names_fo.traverse_entries_fo.traverse_dir_fo.vptr = &traverse_names_traverse_dir_fo_vtbl;
  is_empty_fo->traverse_names_fo.traverse_entries_fo.vptr = &traverse_names_traverse_entries_fo_vtbl;
  is_empty_fo->traverse_names_fo.vptr = &is_empty_traverse_names_fo_vtbl;
  traverse_names_fo_init(&is_empty_fo->traverse_names_fo);
}

/* match */

struct match_fo_struct {
  struct traverse_name_fo_struct traverse_name_fo;
  const char *fs_basename;
  struct longname_reader_struct longname_reader;
};

static void match_do_long_dir_entry(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct match_fo_struct *match_fo;
  const longDirectoryEntryStructure_t *dir_ent;

  match_fo = downcast(struct match_fo_struct, traverse_name_fo, traverse_name_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->long_struct;
  longname_reader_do_long_dir_entry(&match_fo->longname_reader, dir_ent);
}

static int match_do_short_dir_entry(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct match_fo_struct *match_fo;
  const directoryEntryStructure_t *dir_ent;
  const unsigned short *longname;

  match_fo = downcast(struct match_fo_struct, traverse_name_fo, traverse_name_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  longname = longname_reader_do_short_dir_entry(&match_fo->longname_reader, dir_ent);
  return match(match_fo->fs_basename, longname, dir_ent->DIR_Name);
}

static const struct traverse_name_fo_vtbl_struct match_traverse_name_fo_vtbl = {
  .do_long_dir_entry = match_do_long_dir_entry,
  .do_short_dir_entry = match_do_short_dir_entry
};

static void match_fo_init(struct match_fo_struct *match_fo, const char *fs_basename)
{
  match_fo->traverse_name_fo.traverse_entries_fo.traverse_dir_fo.vptr = &traverse_name_traverse_dir_fo_vtbl;
  match_fo->traverse_name_fo.traverse_entries_fo.vptr = &traverse_name_traverse_entries_fo_vtbl;
  match_fo->traverse_name_fo.vptr = &match_traverse_name_fo_vtbl;
  traverse_name_fo_init(&match_fo->traverse_name_fo);
  match_fo->fs_basename = fs_basename;
  longname_reader_init(&match_fo->longname_reader);
}

/* get_name */

struct get_name_fo_struct {
  struct traverse_name_fo_struct traverse_name_fo;
  char *name;
  struct longname_reader_struct longname_reader;
};

static void get_name_do_long_dir_entry(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct get_name_fo_struct *get_name_fo;
  const longDirectoryEntryStructure_t *dir_ent;

  get_name_fo = downcast(struct get_name_fo_struct, traverse_name_fo, traverse_name_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->long_struct;
  longname_reader_do_long_dir_entry(&get_name_fo->longname_reader, dir_ent);
}

static int get_name_do_short_dir_entry(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct get_name_fo_struct *get_name_fo;
  const directoryEntryStructure_t *dir_ent;
  const unsigned short *longname;

  get_name_fo = downcast(struct get_name_fo_struct, traverse_name_fo, traverse_name_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  longname = longname_reader_do_short_dir_entry(&get_name_fo->longname_reader, dir_ent);
  decode(longname, dir_ent->DIR_Name, get_name_fo->name);
  return 0;
}

static const struct traverse_name_fo_vtbl_struct get_name_traverse_name_fo_vtbl = {
  .do_long_dir_entry = get_name_do_long_dir_entry,
  .do_short_dir_entry = get_name_do_short_dir_entry
};

static void get_name_fo_init(struct get_name_fo_struct *get_name_fo, char name[])
{
  get_name_fo->traverse_name_fo.traverse_entries_fo.traverse_dir_fo.vptr = &traverse_name_traverse_dir_fo_vtbl;
  get_name_fo->traverse_name_fo.traverse_entries_fo.vptr = &traverse_name_traverse_entries_fo_vtbl;
  get_name_fo->traverse_name_fo.vptr = &get_name_traverse_name_fo_vtbl;
  traverse_name_fo_init(&get_name_fo->traverse_name_fo);
  get_name_fo->name = name;
  longname_reader_init(&get_name_fo->longname_reader);
}

/* get_attr */

struct get_attr_fo_struct {
  struct traverse_name_fo_struct traverse_name_fo;
  struct vfat_attr_struct *attr;
};

static int get_attr_do_short_dir_entry(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct get_attr_fo_struct *get_attr_fo;
  const directoryEntryStructure_t *dir_ent;

  get_attr_fo = downcast(struct get_attr_fo_struct, traverse_name_fo, traverse_name_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  get_attr_fo->attr->first_cluster = (dir_ent->DIR_FstClusHI << 16) | dir_ent->DIR_FstClusLO;
  get_attr_fo->attr->type = dir_ent->DIR_Attr & ATTR_DIRECTORY ? FT_DIR : FT_REG;
  get_attr_fo->attr->size = dir_ent->DIR_Attr & ATTR_DIRECTORY ? 0 : dir_ent->DIR_FileSize;
  get_attr_fo->attr->flags.read_only = dir_ent->DIR_Attr & ATTR_READ_ONLY ? 1 : 0;
  get_attr_fo->attr->flags.archive = dir_ent->DIR_Attr & ATTR_ARCHIVE ? 1 : 0;
  make_calendar_time(dir_ent->DIR_CrtDate, dir_ent->DIR_CrtTime, dir_ent->DIR_CrtTimeTenth, &get_attr_fo->attr->ctime);
  make_calendar_time(dir_ent->DIR_LstAccDate, 0, 0, &get_attr_fo->attr->adate);
  make_calendar_time(dir_ent->DIR_WrtDate, dir_ent->DIR_WrtTime, 0, &get_attr_fo->attr->mtime);
  return 0;
}

static const struct traverse_name_fo_vtbl_struct get_attr_traverse_name_fo_vtbl = {
  .do_long_dir_entry = traverse_name_skip_long_dir_entry,
  .do_short_dir_entry = get_attr_do_short_dir_entry
};

static void get_attr_fo_init(struct get_attr_fo_struct *get_attr_fo, struct vfat_attr_struct *attr)
{
  get_attr_fo->traverse_name_fo.traverse_entries_fo.traverse_dir_fo.vptr = &traverse_name_traverse_dir_fo_vtbl;
  get_attr_fo->traverse_name_fo.traverse_entries_fo.vptr = &traverse_name_traverse_entries_fo_vtbl;
  get_attr_fo->traverse_name_fo.vptr = &get_attr_traverse_name_fo_vtbl;
  traverse_name_fo_init(&get_attr_fo->traverse_name_fo);
  get_attr_fo->attr = attr;
}

/* set_attr */

struct set_attr_fo_struct {
  struct traverse_name_fo_struct traverse_name_fo;
  const struct vfat_attr_struct *attr;
};

static int set_attr_do_short_dir_entry(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct set_attr_fo_struct *set_attr_fo;
  directoryEntryStructure_t *dir_ent;
  struct calendar_time_struct calendar_time;

  set_attr_fo = downcast(struct set_attr_fo_struct, traverse_name_fo, traverse_name_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  if (set_attr_fo->attr->first_cluster != ((unsigned long)(dir_ent->DIR_FstClusHI << 16) | dir_ent->DIR_FstClusLO)) {
    vfat_dirty_dir_iter_buffer(dir_iter);
    dir_ent->DIR_FstClusHI = set_attr_fo->attr->first_cluster >> 16;
    dir_ent->DIR_FstClusLO = set_attr_fo->attr->first_cluster & 0xffff;
  }
  if (!(dir_ent->DIR_Attr & ATTR_DIRECTORY) && set_attr_fo->attr->size != dir_ent->DIR_FileSize) {
    vfat_dirty_dir_iter_buffer(dir_iter);
    dir_ent->DIR_FileSize = set_attr_fo->attr->size;
  }
  if (set_attr_fo->attr->flags.read_only != (dir_ent->DIR_Attr & ATTR_READ_ONLY) ? 1 : 0) {
    vfat_dirty_dir_iter_buffer(dir_iter);
    if (set_attr_fo->attr->flags.read_only)
      dir_ent->DIR_Attr |= ATTR_READ_ONLY;
    else
      dir_ent->DIR_Attr &= ~ATTR_READ_ONLY;
  }
  if (set_attr_fo->attr->flags.archive != (dir_ent->DIR_Attr & ATTR_ARCHIVE) ? 1 : 0) {
    vfat_dirty_dir_iter_buffer(dir_iter);
    if (set_attr_fo->attr->flags.archive)
      dir_ent->DIR_Attr |= ATTR_ARCHIVE;
    else
      dir_ent->DIR_Attr &= ~ATTR_ARCHIVE;
  }
  make_calendar_time(dir_ent->DIR_CrtDate, dir_ent->DIR_CrtTime, dir_ent->DIR_CrtTimeTenth, &calendar_time);
  if (timecmp(&set_attr_fo->attr->ctime, &calendar_time)) {
    vfat_dirty_dir_iter_buffer(dir_iter);
    dir_ent->DIR_CrtTimeTenth = make_fat_time_tenth(&set_attr_fo->attr->ctime);
    dir_ent->DIR_CrtTime = make_fat_time(&set_attr_fo->attr->ctime);
    dir_ent->DIR_CrtDate = make_fat_date(&set_attr_fo->attr->ctime);
  }
  make_calendar_time(dir_ent->DIR_LstAccDate, 0, 0, &calendar_time);
  if (datecmp(&set_attr_fo->attr->adate, &calendar_time)) {
    vfat_dirty_dir_iter_buffer(dir_iter);
    dir_ent->DIR_LstAccDate = make_fat_date(&set_attr_fo->attr->adate);
  }
  make_calendar_time(dir_ent->DIR_WrtDate, dir_ent->DIR_WrtTime, 0, &calendar_time);
  if (timecmp(&set_attr_fo->attr->mtime, &calendar_time)) {
    vfat_dirty_dir_iter_buffer(dir_iter);
    dir_ent->DIR_WrtTime = make_fat_time(&set_attr_fo->attr->mtime);
    dir_ent->DIR_WrtDate = make_fat_date(&set_attr_fo->attr->mtime);
  }
  return 0;
}

static const struct traverse_name_fo_vtbl_struct set_attr_traverse_name_fo_vtbl = {
  .do_long_dir_entry = traverse_name_skip_long_dir_entry,
  .do_short_dir_entry = set_attr_do_short_dir_entry
};

static void set_attr_fo_init(struct set_attr_fo_struct *set_attr_fo, const struct vfat_attr_struct *attr)
{
  set_attr_fo->traverse_name_fo.traverse_entries_fo.traverse_dir_fo.vptr = &traverse_name_traverse_dir_fo_vtbl;
  set_attr_fo->traverse_name_fo.traverse_entries_fo.vptr = &traverse_name_traverse_entries_fo_vtbl;
  set_attr_fo->traverse_name_fo.vptr = &set_attr_traverse_name_fo_vtbl;
  traverse_name_fo_init(&set_attr_fo->traverse_name_fo);
  set_attr_fo->attr = attr;
}

/* prepare_write_name */
/* Recolte des informations en preparation de write_name :
 * - compte le nombre total d'entrees dans le repertoire ;
 * - verifie l'integrite du repertoire ;
 * - verifie que le nom n'existe pas deja ;
 * - recherche un emplacement libre pour ecrire le nom. */

struct prepare_write_name_fo_struct {
  struct traverse_entries_fo_struct traverse_entries_fo;
  unsigned long *tot_dir_ent_count; /* nombre total d'entrees de repertoire */
  const char *fs_basename;
  unsigned int needed_free_dir_ent_count; /* nombre d'entrees libre necessaires */
  unsigned long *free_dir_ent_key; /* emplacement trouve */
  int errnum;
  int finished; /* plus que des entrees libres */
  struct longname_reader_struct longname_reader;
  int state; /* 0=pas d'emplacement en cours, 1=emplacement en cours, 2=emplacement trouve */
  unsigned int found_free_dir_ent_count; /* nombre d'entrees libre trouvees */
};

static int prepare_write_name_get_retval(struct traverse_dir_fo_struct *traverse_dir_fo)
{
  struct traverse_entries_fo_struct *traverse_entries_fo;
  struct prepare_write_name_fo_struct *prepare_write_name_fo;

  traverse_entries_fo = downcast(struct traverse_entries_fo_struct, traverse_dir_fo, traverse_dir_fo);
  prepare_write_name_fo = downcast(struct prepare_write_name_fo_struct, traverse_entries_fo, traverse_entries_fo);
  if (prepare_write_name_fo->errnum)
    return -prepare_write_name_fo->errnum;
  if (!prepare_write_name_fo->finished)
    return -EINVAL;
  /* On renvoie le nombre d'entrees manquantes. */
  return prepare_write_name_fo->needed_free_dir_ent_count - prepare_write_name_fo->found_free_dir_ent_count;
}

static void prepare_write_name_fo_do_free_entry(struct prepare_write_name_fo_struct *prepare_write_name_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  switch (prepare_write_name_fo->state) {
  case 0: /* premiere entree libre contigue trouvee */
    *prepare_write_name_fo->free_dir_ent_key = vfat_get_dir_iter_dir_ent_key(dir_iter);
    prepare_write_name_fo->found_free_dir_ent_count = 1;
    if (prepare_write_name_fo->found_free_dir_ent_count == prepare_write_name_fo->needed_free_dir_ent_count)
      prepare_write_name_fo->state = 2;
    else
      prepare_write_name_fo->state = 1;
    break;
  case 1: /* on a deja trouve des entrees libres contigues */
    prepare_write_name_fo->found_free_dir_ent_count++;
    if (prepare_write_name_fo->found_free_dir_ent_count == prepare_write_name_fo->needed_free_dir_ent_count)
      prepare_write_name_fo->state = 2;
    break;
  }
}

static int prepare_write_name_do_free_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct prepare_write_name_fo_struct *prepare_write_name_fo;

  prepare_write_name_fo = downcast(struct prepare_write_name_fo_struct, traverse_entries_fo, traverse_entries_fo);
  if (prepare_write_name_fo->finished) {
    prepare_write_name_fo->errnum = EINVAL;
    return 0;
  }
  (*prepare_write_name_fo->tot_dir_ent_count)++;
  longname_reader_do_free_entry(&prepare_write_name_fo->longname_reader);
  prepare_write_name_fo_do_free_entry(prepare_write_name_fo, dir_iter);
  return 1;
}

static int prepare_write_name_do_free_entry_no_more(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct prepare_write_name_fo_struct *prepare_write_name_fo;

  prepare_write_name_fo = downcast(struct prepare_write_name_fo_struct, traverse_entries_fo, traverse_entries_fo);
  if (!prepare_write_name_fo->finished) /* premiere entree nulle */
    if (prepare_write_name_fo->state != 2)
      /* On doit s'assurer qu'il reste au moins une entree de ce type pour
         marquer la fin du repertoire. */
      prepare_write_name_fo->needed_free_dir_ent_count++;
  prepare_write_name_fo->finished = 1;
  (*prepare_write_name_fo->tot_dir_ent_count)++;
  prepare_write_name_fo_do_free_entry(prepare_write_name_fo, dir_iter);
  return 1;
}

static int prepare_write_name_do_long_dir_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct prepare_write_name_fo_struct *prepare_write_name_fo;
  const longDirectoryEntryStructure_t *dir_ent;

  prepare_write_name_fo = downcast(struct prepare_write_name_fo_struct, traverse_entries_fo, traverse_entries_fo);
  if (prepare_write_name_fo->finished) {
    prepare_write_name_fo->errnum = EINVAL;
    return 0;
  }
  (*prepare_write_name_fo->tot_dir_ent_count)++;
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->long_struct;
  longname_reader_do_long_dir_entry(&prepare_write_name_fo->longname_reader, dir_ent);
  switch (prepare_write_name_fo->state) {
  case 1:
    prepare_write_name_fo->state = 0;
    break;
  }
  return 1;
}

static int prepare_write_name_do_short_dir_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct prepare_write_name_fo_struct *prepare_write_name_fo;
  const directoryEntryStructure_t *dir_ent;
  const unsigned short *longname;

  prepare_write_name_fo = downcast(struct prepare_write_name_fo_struct, traverse_entries_fo, traverse_entries_fo);
  if (prepare_write_name_fo->finished) {
    prepare_write_name_fo->errnum = EINVAL;
    return 0;
  }
  (*prepare_write_name_fo->tot_dir_ent_count)++;
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  longname = longname_reader_do_short_dir_entry(&prepare_write_name_fo->longname_reader, dir_ent);
  if (!is_file_dir_entry(dir_ent))
    goto next;
  if (match(prepare_write_name_fo->fs_basename, longname, dir_ent->DIR_Name)) {
    prepare_write_name_fo->errnum = EEXIST;
    return 0;
  }
 next:
  switch (prepare_write_name_fo->state) {
  case 1:
    prepare_write_name_fo->state = 0;
    break;
  }
  return 1;
}

static const struct traverse_dir_fo_vtbl_struct prepare_write_name_traverse_dir_fo_vtbl = {
  .do_dir_entry = traverse_entries_do_dir_entry,
  .get_retval = prepare_write_name_get_retval
};

static const struct traverse_entries_fo_vtbl_struct prepare_write_name_traverse_entries_fo_vtbl = {
  .do_free_entry = prepare_write_name_do_free_entry,
  .do_free_entry_no_more = prepare_write_name_do_free_entry_no_more,
  .do_long_dir_entry = prepare_write_name_do_long_dir_entry,
  .do_short_dir_entry = prepare_write_name_do_short_dir_entry
};

static void prepare_write_name_fo_init(struct prepare_write_name_fo_struct *prepare_write_name_fo, unsigned long *tot_dir_ent_count, const char *fs_basename, unsigned int needed_free_dir_ent_count, unsigned long *free_dir_ent_key)
{
  prepare_write_name_fo->traverse_entries_fo.traverse_dir_fo.vptr = &prepare_write_name_traverse_dir_fo_vtbl;
  prepare_write_name_fo->traverse_entries_fo.vptr = &prepare_write_name_traverse_entries_fo_vtbl;
  prepare_write_name_fo->tot_dir_ent_count = tot_dir_ent_count;
  *prepare_write_name_fo->tot_dir_ent_count = 0;
  prepare_write_name_fo->fs_basename = fs_basename;
  prepare_write_name_fo->needed_free_dir_ent_count = needed_free_dir_ent_count;
  prepare_write_name_fo->free_dir_ent_key = free_dir_ent_key;
  prepare_write_name_fo->errnum = 0;
  prepare_write_name_fo->finished = 0;
  longname_reader_init(&prepare_write_name_fo->longname_reader);
  prepare_write_name_fo->state = 0;
  prepare_write_name_fo->found_free_dir_ent_count = 0;
}

/* write_name */
/* prepare_write_name est suppose avoir ete execute avant. Aucune verification
 * n'est faite sur le repertoire ici. */

struct write_name_fo_struct {
  struct traverse_dir_fo_struct traverse_dir_fo;
  struct longname_writer_struct longname_writer;
  const unsigned char *shortname;
  const struct vfat_attr_struct *attr;
  int done;
};

static int write_name_do_dir_entry(struct traverse_dir_fo_struct *traverse_dir_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct write_name_fo_struct *write_name_fo;
  union vfat_dir_entry_union *dir_ent;

  write_name_fo = downcast(struct write_name_fo_struct, traverse_dir_fo, traverse_dir_fo);
  dir_ent = vfat_get_dir_iter_dir_entry(dir_iter);
  vfat_dirty_dir_iter_buffer(dir_iter);
  if (longname_writer_do_long_dir_entry(&write_name_fo->longname_writer, &dir_ent->long_struct))
    return 1;
  else {
    memcpy(dir_ent->short_struct.DIR_Name, write_name_fo->shortname, 11);
    dir_ent->short_struct.DIR_Attr = (write_name_fo->attr->flags.read_only ? ATTR_READ_ONLY : 0)
      | (write_name_fo->attr->type == FT_DIR ? ATTR_DIRECTORY : 0)
      | (write_name_fo->attr->flags.archive ? ATTR_ARCHIVE : 0);
    dir_ent->short_struct.DIR_NTRes = 0;
    dir_ent->short_struct.DIR_CrtTimeTenth = make_fat_time_tenth(&write_name_fo->attr->ctime);
    dir_ent->short_struct.DIR_CrtTime = make_fat_time(&write_name_fo->attr->ctime);
    dir_ent->short_struct.DIR_CrtDate = make_fat_date(&write_name_fo->attr->ctime);
    dir_ent->short_struct.DIR_LstAccDate = make_fat_date(&write_name_fo->attr->adate);
    dir_ent->short_struct.DIR_FstClusHI = write_name_fo->attr->first_cluster >> 16;
    dir_ent->short_struct.DIR_WrtTime = make_fat_time(&write_name_fo->attr->mtime);
    dir_ent->short_struct.DIR_WrtDate = make_fat_date(&write_name_fo->attr->mtime);
    dir_ent->short_struct.DIR_FstClusLO = write_name_fo->attr->first_cluster & 0xffff;
    dir_ent->short_struct.DIR_FileSize = write_name_fo->attr->size;
    write_name_fo->done = 1;
    return 0;
  }
}

static int write_name_get_retval(struct traverse_dir_fo_struct *traverse_dir_fo)
{
  struct write_name_fo_struct *write_name_fo;

  write_name_fo = downcast(struct write_name_fo_struct, traverse_dir_fo, traverse_dir_fo);
  if (!write_name_fo->done)
    return -EINVAL;
  return 0;
}

static const struct traverse_dir_fo_vtbl_struct write_name_traverse_dir_fo_vtbl = {
  .do_dir_entry = write_name_do_dir_entry,
  .get_retval = write_name_get_retval
};

static void write_name_fo_init(struct write_name_fo_struct *write_name_fo, const unsigned short longname[], const unsigned char shortname[], const struct vfat_attr_struct *attr)
{
  write_name_fo->traverse_dir_fo.vptr = &write_name_traverse_dir_fo_vtbl;
  longname_writer_init(&write_name_fo->longname_writer, longname, chkSum(shortname));
  write_name_fo->shortname = shortname;
  write_name_fo->attr = attr;
  write_name_fo->done = 0;
}

/* clear_name */

struct clear_name_fo_struct {
  struct traverse_name_fo_struct traverse_name_fo;
};

static void clear_name_do_long_dir_entry(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  vfat_dirty_dir_iter_buffer(dir_iter);
  vfat_get_dir_iter_dir_entry(dir_iter)->short_struct.DIR_Name[0] = 0xe5;
}

static int clear_name_do_short_dir_entry(struct traverse_name_fo_struct *traverse_name_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  vfat_dirty_dir_iter_buffer(dir_iter);
  vfat_get_dir_iter_dir_entry(dir_iter)->short_struct.DIR_Name[0] = 0xe5;
  return 0;
}

static const struct traverse_name_fo_vtbl_struct clear_name_traverse_name_fo_vtbl = {
  .do_long_dir_entry = clear_name_do_long_dir_entry,
  .do_short_dir_entry = clear_name_do_short_dir_entry
};

static void clear_name_fo_init(struct clear_name_fo_struct *clear_name_fo)
{
  clear_name_fo->traverse_name_fo.traverse_entries_fo.traverse_dir_fo.vptr = &traverse_name_traverse_dir_fo_vtbl;
  clear_name_fo->traverse_name_fo.traverse_entries_fo.vptr = &traverse_name_traverse_entries_fo_vtbl;
  clear_name_fo->traverse_name_fo.vptr = &clear_name_traverse_name_fo_vtbl;
}

/* write_dot_dotdot */

struct write_dot_dotdot_fo_struct {
  struct traverse_dir_fo_struct traverse_dir_fo;
  unsigned long first_cluster;
  unsigned long parent_first_cluster;
  struct calendar_time_struct ctime;
  struct calendar_time_struct adate;
  struct calendar_time_struct mtime;
  int state;
};

static int write_dot_dotdot_do_dir_entry(struct traverse_dir_fo_struct *traverse_dir_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct write_dot_dotdot_fo_struct *write_dot_dotdot_fo;
  directoryEntryStructure_t *dir_ent;

  write_dot_dotdot_fo = downcast(struct write_dot_dotdot_fo_struct, traverse_dir_fo, traverse_dir_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  vfat_dirty_dir_iter_buffer(dir_iter);
  switch (write_dot_dotdot_fo->state) {
  case 0: /* dot */
    memcpy(dir_ent->DIR_Name, ".          ", 11);
    dir_ent->DIR_Attr = ATTR_DIRECTORY;
    dir_ent->DIR_NTRes = 0;
    dir_ent->DIR_CrtTimeTenth = make_fat_time_tenth(&write_dot_dotdot_fo->ctime);
    dir_ent->DIR_CrtTime = make_fat_time(&write_dot_dotdot_fo->ctime);
    dir_ent->DIR_CrtDate = make_fat_date(&write_dot_dotdot_fo->ctime);
    dir_ent->DIR_LstAccDate = make_fat_date(&write_dot_dotdot_fo->adate);
    dir_ent->DIR_FstClusHI = write_dot_dotdot_fo->first_cluster >> 16;
    dir_ent->DIR_WrtTime = make_fat_time(&write_dot_dotdot_fo->mtime);
    dir_ent->DIR_WrtDate = make_fat_date(&write_dot_dotdot_fo->mtime);
    dir_ent->DIR_FstClusLO = write_dot_dotdot_fo->first_cluster & 0xffff;
    dir_ent->DIR_FileSize = 0;
    write_dot_dotdot_fo->state = 1;
    return 1;
  case 1: /* dotdot */
  default:
    memcpy(dir_ent->DIR_Name, "..         ", 11);
    dir_ent->DIR_Attr = ATTR_DIRECTORY;
    dir_ent->DIR_NTRes = 0;
    dir_ent->DIR_CrtTimeTenth = make_fat_time_tenth(&write_dot_dotdot_fo->ctime);
    dir_ent->DIR_CrtTime = make_fat_time(&write_dot_dotdot_fo->ctime);
    dir_ent->DIR_CrtDate = make_fat_date(&write_dot_dotdot_fo->ctime);
    dir_ent->DIR_LstAccDate = make_fat_date(&write_dot_dotdot_fo->adate);
    dir_ent->DIR_FstClusHI = write_dot_dotdot_fo->parent_first_cluster >> 16;
    dir_ent->DIR_WrtTime = make_fat_time(&write_dot_dotdot_fo->mtime);
    dir_ent->DIR_WrtDate = make_fat_date(&write_dot_dotdot_fo->mtime);
    dir_ent->DIR_FstClusLO = write_dot_dotdot_fo->parent_first_cluster & 0xffff;
    dir_ent->DIR_FileSize = 0;
    write_dot_dotdot_fo->state = 2;
    return 0;
  }
}

static int write_dot_dotdot_get_retval(struct traverse_dir_fo_struct *traverse_dir_fo)
{
  struct write_dot_dotdot_fo_struct *write_dot_dotdot_fo;

  write_dot_dotdot_fo = downcast(struct write_dot_dotdot_fo_struct, traverse_dir_fo, traverse_dir_fo);
  if (write_dot_dotdot_fo->state != 2)
    return -EINVAL;
  return 0;
}

static const struct traverse_dir_fo_vtbl_struct write_dot_dotdot_traverse_dir_fo_vtbl = {
  .do_dir_entry = write_dot_dotdot_do_dir_entry,
  .get_retval = write_dot_dotdot_get_retval
};

static void write_dot_dotdot_fo_init(struct write_dot_dotdot_fo_struct *write_dot_dotdot_fo, unsigned long first_cluster, unsigned long parent_first_cluster, const struct calendar_time_struct *ctime, const struct calendar_time_struct *adate, const struct calendar_time_struct *mtime)
{
  write_dot_dotdot_fo->traverse_dir_fo.vptr = &write_dot_dotdot_traverse_dir_fo_vtbl;
  write_dot_dotdot_fo->first_cluster = first_cluster;
  write_dot_dotdot_fo->parent_first_cluster = parent_first_cluster;
  write_dot_dotdot_fo->ctime = *ctime;
  write_dot_dotdot_fo->adate = *adate;
  write_dot_dotdot_fo->mtime = *mtime;
  write_dot_dotdot_fo->state = 0;
}

/* update_dotdot */

struct update_dotdot_fo_struct {
  struct traverse_entries_fo_struct traverse_entries_fo;
  unsigned long first_cluster;
  int done;
};

static int update_dotdot_get_retval(struct traverse_dir_fo_struct *traverse_dir_fo)
{
  struct traverse_entries_fo_struct *traverse_entries_fo;
  struct update_dotdot_fo_struct *update_dotdot_fo;

  traverse_entries_fo = downcast(struct traverse_entries_fo_struct, traverse_dir_fo, traverse_dir_fo);
  update_dotdot_fo = downcast(struct update_dotdot_fo_struct, traverse_entries_fo, traverse_entries_fo);
  if (!update_dotdot_fo->done)
    return -EINVAL;
  return 0;
}

static int update_dotdot_do_short_dir_entry(struct traverse_entries_fo_struct *traverse_entries_fo, const struct vfat_dir_iter_struct *dir_iter)
{
  struct update_dotdot_fo_struct *update_dotdot_fo;
  directoryEntryStructure_t *dir_ent;

  update_dotdot_fo = downcast(struct update_dotdot_fo_struct, traverse_entries_fo, traverse_entries_fo);
  dir_ent = &vfat_get_dir_iter_dir_entry(dir_iter)->short_struct;
  if (!is_dotdot(dir_ent))
    return 1;
  vfat_dirty_dir_iter_buffer(dir_iter);
  dir_ent->DIR_FstClusHI = update_dotdot_fo->first_cluster >> 16;
  dir_ent->DIR_FstClusLO = update_dotdot_fo->first_cluster & 0xffff;
  update_dotdot_fo->done = 1;
  return 0;
}

static const struct traverse_dir_fo_vtbl_struct update_dotdot_traverse_dir_fo_vtbl = {
  .do_dir_entry = traverse_entries_do_dir_entry,
  .get_retval = update_dotdot_get_retval
};

static const struct traverse_entries_fo_vtbl_struct update_dotdot_traverse_entries_fo_vtbl = {
  .do_free_entry = traverse_entries_skip_free_entry,
  .do_free_entry_no_more = traverse_entries_stop_free_entry_no_more,
  .do_long_dir_entry = traverse_entries_skip_long_dir_entry,
  .do_short_dir_entry = update_dotdot_do_short_dir_entry
};

static void update_dotdot_fo_init(struct update_dotdot_fo_struct *update_dotdot_fo, unsigned long first_cluster)
{
  update_dotdot_fo->traverse_entries_fo.traverse_dir_fo.vptr = &update_dotdot_traverse_dir_fo_vtbl;
  update_dotdot_fo->traverse_entries_fo.vptr = &update_dotdot_traverse_entries_fo_vtbl;
  update_dotdot_fo->first_cluster = first_cluster;
  update_dotdot_fo->done = 0;
}

/* Fonctions-objets pour traverse_file */

/* read */

struct read_fo_struct {
  struct traverse_file_fo_struct traverse_file_fo;
  unsigned long pos;
  char *fs_buf;
  unsigned int count;
  unsigned int n;
};

static int read_do_file_blk(struct traverse_file_fo_struct *traverse_file_fo, const struct vfat_file_iter_struct *file_iter)
{
  const unsigned long file_blk_size = vfat_get_file_iter_file_blk_size(file_iter);

  struct read_fo_struct *read_fo;
  unsigned int size;

  read_fo = downcast(struct read_fo_struct, traverse_file_fo, traverse_file_fo);
  if (read_fo->pos >= file_blk_size) { /* skip */
    read_fo->pos -= file_blk_size;
    return 1;
  }

  size = file_blk_size - read_fo->pos;
  if (size > read_fo->count)
    size = read_fo->count;
  memcpy_tofs(read_fo->fs_buf, vfat_get_file_iter_file_blk_buffer(file_iter) + read_fo->pos, size);
  read_fo->n += size;
  read_fo->fs_buf += size;
  read_fo->count -= size;
  if (!read_fo->count)
    return 0;
  read_fo->pos = 0;
  return 1;
}

static int read_get_retval(struct traverse_file_fo_struct *traverse_file_fo)
{
  struct read_fo_struct *read_fo;

  read_fo = downcast(struct read_fo_struct, traverse_file_fo, traverse_file_fo);
  return read_fo->n;
}

static const struct traverse_file_fo_vtbl_struct read_traverse_file_fo_vtbl = {
  .do_file_blk = read_do_file_blk,
  .get_retval = read_get_retval
};

static void read_fo_init(struct read_fo_struct *read_fo, unsigned long pos, char *fs_buf, unsigned int count)
{
  read_fo->traverse_file_fo.vptr = &read_traverse_file_fo_vtbl;
  read_fo->pos = pos;
  read_fo->fs_buf = fs_buf;
  read_fo->count = count;
  read_fo->n = 0;
}

/* write */

struct write_fo_struct {
  struct traverse_file_fo_struct traverse_file_fo;
  unsigned long pos_zeros;
  unsigned long pos;
  const char *fs_buf;
  unsigned int count;
  unsigned int n;
};

static int write_do_file_blk(struct traverse_file_fo_struct *traverse_file_fo, const struct vfat_file_iter_struct *file_iter)
{
  const unsigned long file_blk_size = vfat_get_file_iter_file_blk_size(file_iter);

  struct write_fo_struct *write_fo;
  char *buf;
  unsigned int size;

  write_fo = downcast(struct write_fo_struct, traverse_file_fo, traverse_file_fo);
  if (write_fo->pos_zeros >= file_blk_size) { /* skip */
    write_fo->pos_zeros -= file_blk_size;
    write_fo->pos -= file_blk_size;
    return 1;
  }
  buf = vfat_get_file_iter_file_blk_buffer(file_iter);
  vfat_dirty_file_iter_buffer(file_iter);
  if (write_fo->pos_zeros < write_fo->pos) {
    if (write_fo->pos >= file_blk_size) {
      memset(buf + write_fo->pos_zeros, 0, file_blk_size - write_fo->pos_zeros);
      write_fo->pos_zeros = 0;
      write_fo->pos -= file_blk_size;
      return 1;
    }
    memset(buf + write_fo->pos_zeros, 0, write_fo->pos - write_fo->pos_zeros);
    write_fo->pos_zeros = 0;
  }
  size = file_blk_size - write_fo->pos;
  if (size > write_fo->count)
    size = write_fo->count;
  memcpy_fromfs(buf + write_fo->pos, write_fo->fs_buf, size);
  write_fo->n += size;
  write_fo->fs_buf += size;
  write_fo->count -= size;
  if (!write_fo->count)
    return 0;
  write_fo->pos = 0;
  return 1;
}

static int write_get_retval(struct traverse_file_fo_struct *traverse_file_fo)
{
  struct write_fo_struct *write_fo;

  write_fo = downcast(struct write_fo_struct, traverse_file_fo, traverse_file_fo);
  return write_fo->n;
}

static const struct traverse_file_fo_vtbl_struct write_traverse_file_fo_vtbl = {
  .do_file_blk = write_do_file_blk,
  .get_retval = write_get_retval
};

static void write_fo_init(struct write_fo_struct *write_fo, unsigned long pos_zeros, unsigned long pos, const char *fs_buf, unsigned int count)
{
  write_fo->traverse_file_fo.vptr = &write_traverse_file_fo_vtbl;
  write_fo->pos_zeros = pos_zeros;
  write_fo->pos = pos;
  write_fo->fs_buf = fs_buf;
  write_fo->count = count;
  write_fo->n = 0;
}

/* Fonctions-enveloppes, exportees */

/* 1 est retourne si le fichier est trouve, 0 si le fichier n'est pas trouve,
 * et un nombre negatif en cas d'erreur. */
int vfat_traverse_find(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, const char *fs_basename, unsigned long *id, int *type)
{
  struct find_fo_struct fo;

  find_fo_init(&fo, fs_basename, id, type);
  return traverse_dir(&fo.traverse_names_fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

/* count doit etre strictement positif. */
int vfat_traverse_read_dir(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, unsigned long pos, char *fs_buf, unsigned int count)
{
  struct read_dir_fo_struct fo;

  read_dir_fo_init(&fo, pos, fs_buf, count);
  return traverse_dir(&fo.traverse_names_fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_collide(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, const unsigned char basename[])
{
  struct collide_fo_struct fo;

  collide_fo_init(&fo, basename);
  return traverse_dir(&fo.traverse_names_fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_clear_dir(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key)
{
  struct clear_dir_fo_struct fo;

  clear_dir_fo_init(&fo);
  return traverse_dir(&fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_is_empty(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key)
{
  struct is_empty_fo_struct fo;

  is_empty_fo_init(&fo);
  return traverse_dir(&fo.traverse_names_fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_match(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, const char *fs_basename)
{
  struct match_fo_struct fo;

  match_fo_init(&fo, fs_basename);
  return traverse_dir(&fo.traverse_name_fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_get_name(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, char name[])
{
  struct get_name_fo_struct fo;

  get_name_fo_init(&fo, name);
  return traverse_dir(&fo.traverse_name_fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_get_attr(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, struct vfat_attr_struct *attr)
{
  struct get_attr_fo_struct fo;

  get_attr_fo_init(&fo, attr);
  return traverse_dir(&fo.traverse_name_fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_set_attr(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, const struct vfat_attr_struct *attr)
{
  struct set_attr_fo_struct fo;

  set_attr_fo_init(&fo, attr);
  return traverse_dir(&fo.traverse_name_fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_prepare_write_name(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, unsigned long *tot_dir_ent_count, const char *fs_basename, unsigned int needed_free_dir_ent_count, unsigned long *free_dir_ent_key)
{
  struct prepare_write_name_fo_struct fo;

  prepare_write_name_fo_init(&fo, tot_dir_ent_count, fs_basename, needed_free_dir_ent_count, free_dir_ent_key);
  return traverse_dir(&fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_write_name(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, const unsigned short longname[], const unsigned char shortname[], const struct vfat_attr_struct *attr)
{
  struct write_name_fo_struct fo;

  write_name_fo_init(&fo, longname, shortname, attr);
  return traverse_dir(&fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_clear_name(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key)
{
  struct clear_name_fo_struct fo;

  clear_name_fo_init(&fo);
  return traverse_dir(&fo.traverse_name_fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_write_dot_dotdot(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, unsigned long first_cluster, unsigned long parent_first_cluster, const struct calendar_time_struct *ctime, const struct calendar_time_struct *adate, const struct calendar_time_struct *mtime)
{
  struct write_dot_dotdot_fo_struct fo;

  write_dot_dotdot_fo_init(&fo, first_cluster, parent_first_cluster, ctime, adate, mtime);
  return traverse_dir(&fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

int vfat_traverse_update_dotdot(struct vfat_instance_struct *vfat_instance, unsigned long dir_ent_key, unsigned long first_cluster)
{
  struct update_dotdot_fo_struct fo;

  update_dotdot_fo_init(&fo, first_cluster);
  return traverse_dir(&fo.traverse_entries_fo.traverse_dir_fo, vfat_instance, dir_ent_key);
}

/* count doit etre strictement positif. */
int vfat_traverse_read(struct vfat_instance_struct *vfat_instance, unsigned long file_blk_key, unsigned long pos, char *fs_buf, unsigned int count)
{
  struct read_fo_struct fo;

  read_fo_init(&fo, pos, fs_buf, count);
  return traverse_file(&fo.traverse_file_fo, vfat_instance, file_blk_key);
}

/* pos_zeros doit etre inferieur ou egal a pos.
 * count doit etre strictement positif. */
int vfat_traverse_write(struct vfat_instance_struct *vfat_instance, unsigned long file_blk_key, unsigned long pos_zeros, unsigned long pos, const char *fs_buf, unsigned int count)
{
  struct write_fo_struct fo;

  write_fo_init(&fo, pos_zeros, pos, fs_buf, count);
  return traverse_file(&fo.traverse_file_fo, vfat_instance, file_blk_key);
}
