/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <it.h>
#include <fd.h>
#include <fs.h>
#include <buffer.h>
#include <device_manager.h>
#include <rtc.h>
#include <rtc_structs.h>
#include <floppy.h>
#include <console.h>
#include <keyboard.h>
#include <video.h>
#include <beep.h>
#include <timer.h>
#include <pic.h>
#include <mm.h>
#include <paging.h>
#include <proc.h>
#include <panic.h>
#include <printk.h>
#include <time.h>
#include <idt.h>
#include <gdt.h>
#include <asm.h>
#include <consts.h>
#include <config.h>

static struct {
  unsigned floppy_drive : 1;
} detected_devices;

static void setup()
{
  setup_gdt();
  setup_idt();
  setup_paging();
}

static void set_irq(unsigned int irq, void *handler)
{
  idt_set_interrupt_gate(32 + irq, handler);
  pic_enable_irq(irq);
}

static void init_console()
{
  video_init();
  kb_init();
  console_init();
  /* Le caractere \a (sonnerie) ne doit pas etre utilise avant l'activation des
     interruptions. */
}

static void init_keyboard()
{
  /* ... en realite le pilote de clavier a ete initialise dans
     init_console(). */
  printk("Keyboard driver initialized.\n");
  kb_print_info();
  set_irq(1, keyboard_it);
  printk("Keyboard interrupt on IRQ 1\n");
}

static void init_timer()
{
  printk("Initializing timer driver... ");
  timer_init();
  printk("done.\n");
  timer_print_info();
  set_irq(0, timer_it);
  printk("Timer interrupt on IRQ 0\n");
}

static void init_floppy()
{
  printk("Initializing floppy disk controller driver... ");
  if (!floppy_init()) {
    printk("no floppy drive detected.\n");
    detected_devices.floppy_drive = 0;
    return;
  }
  detected_devices.floppy_drive = 1;
  printk("done.\n");
  floppy_print_info();
  set_irq(6, floppy_it);
  printk("Floppy interrupt on IRQ 6\n");
}

static void init_device_manager()
{
  device_manager_init();
  if (detected_devices.floppy_drive)
    device_set_present(DEV_FD);
}

/* Affiche un message de bienvenue :-) */
static void print_welcome_msg()
{
  printk("Welcome to " OS_NAME "!\n");
  printk("Version: " OS_VERSION ", built: " OS_BUILD "\n");
}

/* Affiche des informations sur la configuration du systeme. */
static void print_info()
{
  static const char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
  static const char *diskette_drive_type[] = {"no drive", "360KB", "1.2MB", "720KB", "1.44MB"};
  static const char *display[] = {"EGA/VGA", "40x25 colour", "80x25 colour", "80x25 monochrome"};
  static const char *coprocessor[] = {"no", "yes"};

  struct rtc_clock_struct clock;
  struct rtc_configuration_struct configuration;

  /* lecture de la RTC et des donnees de configuration du CMOS */
  rtc_get_clock(&clock);
  rtc_get_configuration(&configuration);

  printk("Real Time Clock: %s %d %d %02d:%02d:%02d\n", months[clock.mon - 1], clock.mday, clock.year, clock.hour, clock.min, clock.sec);
  printk("Diskette drive 0 type: %s\n", diskette_drive_type[configuration.diskette_drive_0_type]);
  printk("Primary display: %s\n", display[configuration.primary_display]);
  printk("Math coprocessor: %s\n", coprocessor[configuration.math_coprocessor]);
  printk("Base memory: %huKB\n", configuration.base_memory);
  /* Note : configuration.extended_memory peut etre faux si il y a plus de
     64 Mo de memoire, cf
     http://www.nondot.org/sabre/os/files/Misc/os-faq/os-faq-memory.html#determine_memory */
}

static void set_exception_handlers()
{
  unsigned int i;

  idt_set_trap_gate(0, divide_error_it); /* Divide error */
  idt_set_trap_gate(1, debug_it); /* Debug exceptions */
  idt_set_interrupt_gate(2, nonmaskable_interrupt_it); /* Nonmaskable interrupt */
  idt_set_system_gate(3, breakpoint_it); /* Breakpoint (one-byte INT 3 instruction) */
  idt_set_system_gate(4, overflow_it); /* Overflow (INTO instruction) */
  idt_set_system_gate(5, bounds_check_it); /* Bounds check (BOUND instruction) */
  idt_set_trap_gate(6, invalid_opcode_it); /* Invalid opcode */
  idt_set_trap_gate(7, device_not_available_it); /* Coprocessor not available */
  idt_set_trap_gate(8, double_fault_it); /* Double fault */
  idt_set_trap_gate(9, coprocessor_segment_overrun_it); /* Coprocessor segment overrun */
  idt_set_trap_gate(10, invalid_tss_it); /* Invalid TSS */
  idt_set_trap_gate(11, segment_not_present_it); /* Segment not present */
  idt_set_trap_gate(12, stack_segment_fault_it); /* Stack exception */
  idt_set_trap_gate(13, general_protection_it); /* General protection */
  idt_set_trap_gate(14, page_fault_it); /* Page fault */
  idt_set_trap_gate(15, ignore_it); /* (reserved) */
  idt_set_trap_gate(16, coprocessor_error_it); /* Coprecessor error */
  idt_set_trap_gate(17, alignment_check_it); /* Alignement check (486) */
  idt_set_trap_gate(18, machine_check_it); /* Machine check (Pentium) */
  for (i = 19; i <= 31; i++)
    idt_set_trap_gate(i, ignore_it); /* (reserved) */
}

static void init()
{
  pic_init();
  time_init();
  init_console();
  set_exception_handlers();
  print_welcome_msg();
  print_info();
  paging_print_info();
  init_keyboard();
  init_timer();
  beep_init();
  init_floppy();
  mm_init();
  init_device_manager();
  buffer_init();
  buffer_print_info();
  fs_init();
  fs_print_info();
  fd_init();
  fd_print_info();
  proc_init();
  proc_print_info();
  idt_set_system_gate(128, syscall_it); /* appels systemes */
}

void __attribute__ ((noreturn)) main()
{
  /* initialisation des structures du processeur */
  setup();

  /* initialisation des pilotes et du noyau */
  init();

  /* lancement de init */
  printk("Starting init...\n");
  if (spawn_init() < 0)
    panic("spawn_init() failed in main()!\n");
  sti();

  /* un systeme d'exploitation ne retourne jamais ;-) */
  while (1)
    hlt();
}
