/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "exec_reader.h"
#include "auxiliary_vector.h"
#include "sched.h"
#include "proc_struct.h"

#include <resolve.h>
#include <node.h>
#include <stat_struct.h>
#include <vm.h>
#include <kmalloc.h>
#include <verify_area.h>
#include <segment.h>
#include <errno.h>
#include <consts.h>
#include <enums.h>
#include <i386.h>
#include <util.h>
#include <string.h>
#include <limits.h>
#include <elf_sysv4.h>

struct elf_reader_struct;

static int load_interpreter(const char *fs_filename, unsigned long *base_address, unsigned long *entry_point);

static inline int check_for_addition(unsigned long base, unsigned long offset)
{
  return base <= ULONG_MAX - offset;
}

/* vaddr doit etre non aligne (car verify_area() doit avoir une longueur non
 * nulle). */
static int pad_page(unsigned long vaddr)
{
  unsigned long size;
  int rv;

  size = PAGE_ALIGN(vaddr) - vaddr;
  if ((rv = verify_area((void *)vaddr, size, PF_WRITE)) < 0)
    return rv;
  memset_tofs((void *)vaddr, 0, size);
  return 0;
}

static int load_segment(struct file_struct *file, unsigned long offset, unsigned long vaddr, unsigned long file_size, unsigned long mem_size, int prot, unsigned long base_address, int main)
{
  unsigned long poff, map_size;
  int retval;

  /* projection du segment du fichier en memoire */
  if (!file_size)
    return 0;
  poff = PAGE_OFFSET(offset);
  offset -= poff;
  vaddr -= poff;
  file_size += poff;
  mem_size += poff;
  if ((retval = vm_map_file(&current->vm, base_address + vaddr, PAGE_ALIGN(file_size), prot, file, offset)) < 0)
    return retval;
  if (prot == PROT_RDWR && PAGE_OFFSET(file_size))
    if ((retval = pad_page(base_address + vaddr + file_size)) < 0)
      return retval;

  /* completion par une projection anonyme */
  map_size = PAGE_ALIGN(file_size);
  if (mem_size > map_size) {
    vaddr += map_size;
    mem_size -= map_size;
    if ((retval = vm_map_anon(&current->vm, base_address + vaddr, PAGE_ALIGN(mem_size), prot, 0, main ? "heap" : NULL)) < 0)
      return retval;
  }

  /* mise a jour de l'adresse de l'interruption de programme */
  if (main)
    set_if_lower(current->brk, base_address + vaddr + mem_size);

  return 0;
}

static const unsigned char elf_magic_number[4] = {ELFMAG0, ELFMAG1, ELFMAG2, ELFMAG3};

struct elf_reader_struct {
  struct exec_reader_struct exec_reader;
  struct file_struct *file;
  Elf32_Ehdr ehdr;
  unsigned long base_address;
};

/* Destructeur. */
static void elf_reader_destroy(struct elf_reader_struct *elf_reader)
{
  node_release_write_access(elf_reader->file);
  node_release(elf_reader->file);
}

static int load_loadable_segment(struct elf_reader_struct *elf_reader, const Elf32_Phdr *program_header, int main)
{
  /* verification de la validite de l'en-tete de programme */
  if (PAGE_OFFSET(program_header->p_vaddr) != PAGE_OFFSET(program_header->p_offset))
    return -EINVAL;
  if (program_header->p_memsz < program_header->p_filesz)
    return -EINVAL;
  if (!check_for_addition(program_header->p_memsz, PAGE_SIZE - 1))
    return -EINVAL;
  if (!check_for_addition(program_header->p_vaddr, program_header->p_memsz))
    return -EINVAL;

  /* chargement du segment en memoire */
  return load_segment(elf_reader->file, program_header->p_offset, program_header->p_vaddr, program_header->p_filesz, program_header->p_memsz, program_header->p_flags & PF_W ? PROT_RDWR : PROT_RDONLY, elf_reader->base_address, main);
}

static int load_library(struct elf_reader_struct *elf_reader)
{
  Elf32_Phdr *program_header_table;
  const Elf32_Phdr *program_header;
  unsigned long total_size;
  int retval;
  int i;

  if (elf_reader->ehdr.e_phentsize != sizeof (Elf32_Phdr))
    return -EINVAL;
  if (!(program_header_table = kmalloc(elf_reader->ehdr.e_phnum * sizeof (Elf32_Phdr))))
    return -ENOMEM;
  if ((retval = read_node(elf_reader->file, elf_reader->ehdr.e_phoff, program_header_table, elf_reader->ehdr.e_phnum * sizeof (Elf32_Phdr))) < 0)
    goto error_kfree_program_header_table;
  switch (elf_reader->ehdr.e_type) {
  case ET_EXEC:
    elf_reader->base_address = 0;
    break;
  case ET_DYN:
    total_size = 0;
    for (i = 0; i < elf_reader->ehdr.e_phnum; i++) {
      program_header = &program_header_table[i];
      switch (program_header->p_type) {
      case PT_LOAD:
        set_if_lower(total_size, PAGE_ALIGN(program_header->p_vaddr + program_header->p_memsz));
        break;
      }
    }
    if (!(elf_reader->base_address = vm_get_free_area(&current->vm, 0, MAPS_TOP, total_size, 0))) {
      retval = -ENOMEM;
      goto error_kfree_program_header_table;
    }
    break;
  }
  for (i = 0; i < elf_reader->ehdr.e_phnum; i++) {
    program_header = &program_header_table[i];
    switch (program_header->p_type) {
    case PT_LOAD:
      if ((retval = load_loadable_segment(elf_reader, program_header, 0)) < 0)
        goto error_kfree_program_header_table;
      break;
    }
  }
  kfree(program_header_table, elf_reader->ehdr.e_phnum * sizeof (Elf32_Phdr));
  return 0;

 error_kfree_program_header_table:
  kfree(program_header_table, elf_reader->ehdr.e_phnum * sizeof (Elf32_Phdr));
  return retval;
}

static int load_program(struct elf_reader_struct *elf_reader, unsigned long *entry_point, struct auxiliary_vector_struct *auxiliary_vector)
{
  Elf32_Phdr *program_header_table;
  struct {
    unsigned interpreter : 1;
    unsigned program_header_table : 1;
  } contents;
  unsigned long interpreter_filename_addr;
  unsigned long program_header_table_addr;
  unsigned long interpreter_base_address;
  unsigned long interpreter_entry_point;
  const Elf32_Phdr *program_header;
  int retval;
  int i;

  switch (elf_reader->ehdr.e_type) {
  case ET_EXEC:
    elf_reader->base_address = 0;
    break;
  case ET_DYN:
    elf_reader->base_address = PIE_BASE_ADDR;
    break;
  }
  if (elf_reader->ehdr.e_phentsize != sizeof (Elf32_Phdr))
    return -EINVAL;
  if (!(program_header_table = kmalloc(elf_reader->ehdr.e_phnum * sizeof (Elf32_Phdr))))
    return -ENOMEM;
  if ((retval = read_node(elf_reader->file, elf_reader->ehdr.e_phoff, program_header_table, elf_reader->ehdr.e_phnum * sizeof (Elf32_Phdr))) < 0)
    goto error_kfree_program_header_table;
  contents.interpreter = 0;
  contents.program_header_table = 0;
  interpreter_filename_addr = 0;
  interpreter_base_address = 0;
  interpreter_entry_point = 0;
  program_header_table_addr = 0;
  for (i = 0; i < elf_reader->ehdr.e_phnum; i++) {
    program_header = &program_header_table[i];
    switch (program_header->p_type) {
    case PT_LOAD:
      if ((retval = load_loadable_segment(elf_reader, program_header, 1)) < 0)
        goto error_kfree_program_header_table;
      if (!contents.program_header_table && elf_reader->ehdr.e_phoff >= program_header->p_offset && elf_reader->ehdr.e_phoff + elf_reader->ehdr.e_phnum * sizeof (Elf32_Phdr) <= program_header->p_offset + program_header->p_filesz) {
        contents.program_header_table = 1;
        program_header_table_addr = elf_reader->base_address + program_header->p_vaddr - program_header->p_offset + elf_reader->ehdr.e_phoff;
      }
      break;
    case PT_INTERP:
      contents.interpreter = 1;
      interpreter_filename_addr = elf_reader->base_address + program_header->p_vaddr;
      break;
    case PT_PHDR:
      contents.program_header_table = 1;
      program_header_table_addr = elf_reader->base_address + program_header->p_vaddr;
      break;
    }
  }
  kfree(program_header_table, elf_reader->ehdr.e_phnum * sizeof (Elf32_Phdr));
  if (contents.interpreter)
    if ((retval = load_interpreter((char *)interpreter_filename_addr, &interpreter_base_address, &interpreter_entry_point)) < 0)
      return retval;

  if (contents.interpreter)
    *entry_point = interpreter_entry_point;
  else
    *entry_point = elf_reader->base_address + elf_reader->ehdr.e_entry;
  if (contents.program_header_table) {
    auxiliary_vector->contents.program_header_table = 1;
    auxiliary_vector->program_header_table = program_header_table_addr;
  }
  else
    auxiliary_vector->contents.program_header_table = 0;
  auxiliary_vector->contents.program_header_entry_size = 1;
  auxiliary_vector->program_header_entry_size = sizeof (Elf32_Phdr);
  auxiliary_vector->contents.program_header_entry_number = 1;
  auxiliary_vector->program_header_entry_number = elf_reader->ehdr.e_phnum;
  auxiliary_vector->contents.page_size = 1;
  auxiliary_vector->page_size = PAGE_SIZE;
  if (contents.interpreter) {
    auxiliary_vector->contents.interpreter_base_address = 1;
    auxiliary_vector->interpreter_base_address = interpreter_base_address;
  }
  else
    auxiliary_vector->contents.interpreter_base_address = 0;
  auxiliary_vector->contents.program_entry_point = 1;
  auxiliary_vector->program_entry_point = elf_reader->base_address + elf_reader->ehdr.e_entry;
  auxiliary_vector->contents.not_elf = 0;
  auxiliary_vector->contents.vdso_entry_point = 0;
  auxiliary_vector->contents.vdso_elf_header = 0;
  return 0;

 error_kfree_program_header_table:
  kfree(program_header_table, elf_reader->ehdr.e_phnum * sizeof (Elf32_Phdr));
  return retval;
}

/* Fonctions virtuelles */

static int elf_exec_reader_load(struct exec_reader_struct *exec_reader, unsigned long *entry_point, struct auxiliary_vector_struct *auxiliary_vector)
{
  return load_program(downcast(struct elf_reader_struct, exec_reader, exec_reader), entry_point, auxiliary_vector);
}

static void elf_exec_reader_free(struct exec_reader_struct *exec_reader)
{
  struct elf_reader_struct *elf_reader;

  elf_reader = downcast(struct elf_reader_struct, exec_reader, exec_reader);
  elf_reader_destroy(elf_reader);
  kfree(elf_reader, sizeof (struct elf_reader_struct));
}

static const struct exec_reader_vtbl_struct elf_exec_reader_vtbl = {
  .load = elf_exec_reader_load,
  .free = elf_exec_reader_free
};

/* Constructeur. */
static int elf_reader_init(struct elf_reader_struct *elf_reader, struct file_struct *file)
{
  int retval;

  elf_reader->exec_reader.vptr = &elf_exec_reader_vtbl;
  elf_reader->file = file;
  if ((retval = read_node(file, 0, (char *)&elf_reader->ehdr, sizeof (Elf32_Ehdr))) < 0)
    return retval;
  if ((unsigned int)retval < sizeof (Elf32_Ehdr))
    return -EINVAL;
  if (elf_reader->ehdr.e_ident[EI_CLASS] != ELFCLASS32
      || elf_reader->ehdr.e_ident[EI_DATA] != ELFDATA2LSB
      || elf_reader->ehdr.e_ident[EI_VERSION] != EV_CURRENT
      || elf_reader->ehdr.e_machine != EM_386
      || elf_reader->ehdr.e_version != EV_CURRENT)
    return -EINVAL;
  if (elf_reader->ehdr.e_type != ET_EXEC && elf_reader->ehdr.e_type != ET_DYN)
    return -ENOEXEC;
  return 0;
}

/* Fonctions exportees */

int is_elf(struct file_struct *file)
{
  unsigned char buf[4];
  int rv;

  if ((rv = read_node(file, 0, (char *)buf, 4)) < 0)
    return rv;
  if (rv < 4)
    return 0;
  return memcmp(buf, elf_magic_number, 4) ? 0 : 1;
}

int alloc_elf_exec_reader(struct file_struct *file, struct exec_reader_struct **exec_reader)
{
  struct elf_reader_struct *elf_reader;
  int retval;

  if (!(elf_reader = kmalloc(sizeof (struct elf_reader_struct))))
    return -ENOMEM;
  if ((retval = elf_reader_init(elf_reader, file)) < 0)
    goto error_kfree;
  *exec_reader = &elf_reader->exec_reader;
  return 0;

 error_kfree:
  kfree(elf_reader, sizeof (struct elf_reader_struct));
  return retval;
}

/* Interpreteur de programme */

static int load_interpreter(const char *fs_filename, unsigned long *base_address, unsigned long *entry_point)
{
  struct file_struct *file;
  struct stat_struct statbuf;
  struct elf_reader_struct elf_reader;
  int retval;

  if ((retval = resolve(fs_filename, &file)) < 0)
    return retval;
  if ((retval = stat_node(file, &statbuf)) < 0)
    goto error_release;
  if (statbuf.type != FT_REG) {
    retval = -EACCES;
    goto error_release;
  }
  if ((retval = node_deny_write_access(file)) < 0)
    goto error_release;
  if ((retval = is_elf(file)) < 0)
    goto error_release_write_access;
  if (!retval) {
    retval = -ELIBBAD;
    goto error_release_write_access;
  }
  if ((retval = elf_reader_init(&elf_reader, file)) < 0)
    goto error_release_write_access;
  if ((retval = load_library(&elf_reader)) < 0)
    goto error_release_write_access;
  *base_address = elf_reader.base_address;
  *entry_point = elf_reader.base_address + elf_reader.ehdr.e_entry;
  elf_reader_destroy(&elf_reader);
  return 0;

 error_release_write_access:
  node_release_write_access(file);
 error_release:
  node_release(file);
  return retval;
}
