/* Global Descriptor Table */
/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/* La GDT provisoire mise en place par le code de demarrage peut etre utilisee
 * sereinement le temps de la mise en place de la GDT definitive. */

/* La GDT definit les descripteurs de segments pour le mode noyau et le mode
 * utilisateur. Les selecteurs correspondants sont fournis sous forme de macros
 * par include/gdt.h. Ils sont egalement utilises dans kernel/it.S. */

/* Structure de la GDT :
 *   0     : inutilise
 *   1     : segment de code noyau
 *   2     : segment de donnees noyau
 *   3     : segment de code utilisateur
 *   4     : segment de donnees utilisateur
 *   5 + n : TSS(n)
 */

#include <system.h>
#include <system_data_structs.h>
#include <system_structs.h>
#include <asm.h>
#include <config.h>

static struct descr_struct gdt[5 + NR_TASKS];

/* Descripteurs */

static void set_code_seg_descr(unsigned int index, int dpl, unsigned long base, unsigned long limit)
{
  struct descr_data_struct sd_dat;

  sd_dat.type = 1; /* segment */
  sd_dat.accessed = 0;
  sd_dat.segment_type = 1; /* code */
  sd_dat.readable = 1;
  sd_dat.conforming = 0;
  sd_dat.dpl = dpl;
  sd_dat.present = 1;
  sd_dat.base = base;
  sd_dat.limit = limit;
  sd_dat.granularity = 1;
  sd_dat.avl = 0;
  sd_dat.def = 1; /* 32-bit */
  set_descr(&gdt[index], &sd_dat);
}

static void set_data_seg_descr(unsigned int index, int dpl, unsigned long base, unsigned long limit)
{
  struct descr_data_struct sd_dat;

  sd_dat.type = 1; /* segment */
  sd_dat.accessed = 0;
  sd_dat.segment_type = 0; /* data */
  sd_dat.writable = 1;
  sd_dat.expand_down = 0;
  sd_dat.dpl = dpl;
  sd_dat.present = 1;
  sd_dat.base = base;
  sd_dat.limit = limit;
  sd_dat.granularity = 1;
  sd_dat.avl = 0;
  sd_dat.big = 1; /* 32-bit */
  set_descr(&gdt[index], &sd_dat);
}

static void set_tss_descr(unsigned int index, const struct tss_struct *tss)
{
  struct descr_data_struct td_dat;

  td_dat.type = 0; /* system */
  td_dat.system_type = SDT_AVL386TSS; /* TSS */
  td_dat.dpl = 0;
  td_dat.present = 1;
  td_dat.base = (unsigned long)tss;
  td_dat.limit = sizeof (struct tss_struct) - 1;
  td_dat.granularity = 0;
  td_dat.avl = 0;
  td_dat.x = 0;
  set_descr(&gdt[index], &td_dat);
}

/* Fonctions exportees */

void setup_gdt()
{
  struct __attribute__ ((packed)) {
    unsigned short limit;
    unsigned long base;
  } gdt_descr;

  set_code_seg_descr(1, 0, 0, 0xfffff); /* segment de code noyau */
  set_data_seg_descr(2, 0, 0, 0xfffff); /* segment de donnees noyau */

  gdt_descr.limit = sizeof gdt - 1;
  gdt_descr.base = (unsigned long)gdt;
  lgdt(gdt_descr);

  /* Bien que la nouvelle GDT soit parfaitement compatible avec les
     selecteurs de la GDT provisoire, il est de bon gout de recharger les
     registres de segments, afin de recharger egalement les descripteurs dans
     les registres internes. */
  load_cs(0x8);
  load_ds(0x10);
  load_es(0x10);
  load_fs(0x10);
  load_gs(0x10);
  load_ss(0x10);
}

void gdt_set_user_code_seg_descr(unsigned long base, unsigned long limit)
{
  set_code_seg_descr(3, 3, base, limit);
}

void gdt_set_user_data_seg_descr(unsigned long base, unsigned long limit)
{
  set_data_seg_descr(4, 3, base, limit);
}

void gdt_set_tss_descr(unsigned int task_nr, const struct tss_struct *tss)
{
  set_tss_descr(5 + task_nr, tss);
}

unsigned short get_tss_sel(unsigned int task_nr)
{
  struct sel_data_struct sel_data;
  unsigned short tss_sel;

  sel_data.rpl = 0; /* mode noyau */
  sel_data.table_ind = 0; /* GDT */
  sel_data.index = 5 + task_nr; /* TSS */
  set_sel(&tss_sel, &sel_data);
  return tss_sel;
}

unsigned int get_task_nr(unsigned short tss_sel)
{
  struct sel_data_struct sel_data;

  get_sel(&tss_sel, &sel_data);
  return sel_data.index - 5;
}
