/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <rtc.h>
#include <rtc_structs.h>
#include <time_struct.h>
#include <config.h>

long startup_time;

struct dst_date_struct {
  int mon;                    /* mois (0-11) */
  int mweek;                  /* semaine du mois (1-5) */
  int wday;                   /* jour de la semaine (0-6) */
  long time;                  /* heure en secondes */
};

struct dst_struct {
  long offset;                /* decalage en secondes */
  struct dst_date_struct start; /* date de debut de l'heure d'ete */
  struct dst_date_struct end; /* date de retour a l'heure d'hiver */
};

struct tz_struct {
  long offset;                /* decalage en secondes */
  const struct dst_struct *dst; /* heure d'ete */
};

static const struct dst_struct dst = {
  TZ_DST_OFFSET,
  {TZ_DST_START_MON, TZ_DST_START_MWEEK, TZ_DST_START_WDAY, TZ_DST_START_TIME},
  {TZ_DST_END_MON, TZ_DST_END_MWEEK, TZ_DST_END_WDAY, TZ_DST_END_TIME}
};
static const struct tz_struct tz = {TZ_STD_OFFSET, &dst};

static const int mdays_common[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
static const int mdays_leap[12] = {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

static const int month_start_common[12] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};
static const int month_start_leap[12] = {0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335};

#define is_leap(year) ((((year) % 4 == 0 && (year) % 100 != 0)) || ((year) % 400 == 0))
#define ydays(year)   (is_leap(year) ? 366 : 365)

#define mdays(year, mon)  (is_leap(year) ? mdays_leap[mon] : mdays_common[mon])
#define month_start(year) (is_leap(year) ? month_start_leap : month_start_common)

#define leap_years(year, n) (((n) + 1) * ((year) - 1) / (n) - (year))
#define year_wday(year)     ((1 + (year) + (leap_years((year), 4) - leap_years((year), 100) + leap_years((year), 400))) % 7)
#define wday(year, yday)    ((year_wday(year) + (yday)) % 7)

#define dsecs 86400

/* Retourne le nombre de secondes depuis le debut de l'annee. */
static long ysec(const struct dst_date_struct *dst_date, int year)
{
  int mon;
  int mon_yday;
  long msec;
  int wday;
  int mday;
  int mweek;
  int i, mdays;

  mon = dst_date->mon;               /* mois */
  mon_yday = month_start(year)[mon]; /* jour de l'annee du debut du mois */

  msec = 0;                    /* nombre de secondes depuis le debut du mois */
  wday = wday(year, mon_yday); /* jour de la semaine */
  mday = 0;                    /* jour du mois */
  mweek = 0;                   /* semaine du mois */
  for (i = 0, mdays = mdays(year, mon); i < mdays; i++) { /* parcours du mois */
    if (wday == dst_date->wday) {
      mweek++;
      mday = i;
    }
    if (mweek == dst_date->mweek)
      break;
    msec += dsecs;
    wday = (wday + 1) % 7;
  }
  return (mon_yday + mday) * dsecs + dst_date->time; /* mois + jour + heure */
}

/* Dit si une date UNIX est dans la periode d'heure d'ete d'un fuseau horaire. */
static int is_dst(long time)
{
  long t;
  int year;
  long dst_start, dst_end;
  long ysecs;

  if (!tz.dst)
    return 0;

  t = 0;
  year = 1970;
  if (time < 0)
    do
      year--, t -= ydays(year) * dsecs;
    while (time < t);
  else
    while (time >= t + (ysecs = ydays(year) * dsecs))
      t += ysecs, year++;
  dst_start = t + tz.offset + ysec(&tz.dst->start, year);
  dst_end = t + tz.dst->offset + ysec(&tz.dst->end, year);
  return dst_start < dst_end
    ? time >= dst_start && time < dst_end
    : time < dst_start || time >= dst_end;
}

/* Retourne l'heure de l'ordinateur sous forme calendrier.
 * Cette heure peut etre locale ou universelle selon la configuration de
 * l'horloge materielle de l'ordinateur. */
void get_calendar_time(struct calendar_time_struct *calendar_time)
{
  struct rtc_clock_struct clock;

  rtc_get_clock(&clock);
  calendar_time->sec = clock.sec;
  calendar_time->min = clock.min;
  calendar_time->hour = clock.hour;
  calendar_time->mday = clock.mday;
  calendar_time->mon = clock.mon;
  calendar_time->year = clock.year;
}

#define normalize(var1, max, var2) ({ \
  register int _tmp; \
  if (var1 < 0) { \
    _tmp = (max - var1) / max; \
    var2 -= _tmp; \
    var1 += _tmp * max; \
  } \
  else if (var1 >= max) { \
    _tmp = var1 / max; \
    var2 += _tmp; \
    var1 -= _tmp * max; \
  } \
})

/* Convertit une heure sous forme calendrier en heure systeme. L'heure sous
 * forme calendrier est consideree comme locale ou universelle selon la
 * configuration de l'horloge materielle de l'ordinateur. */
long make_time(const struct calendar_time_struct *calendar_time)
{
  int sec;   /* 0-59 */
  int min;   /* 0-59 */
  int hour;  /* 0-23 */
  int mday;  /* 0-30 */
  int mon;   /* 0-11 */
  int year;
  int yday;  /* 0-365 */
  int isdst;
  int mdays;
  long t;
  int y;

  sec = calendar_time->sec;
  min = calendar_time->min;
  hour = calendar_time->hour;
  mday = calendar_time->mday - 1;
  mon = calendar_time->mon - 1;
  year = calendar_time->year;

  /* ajustements */
  normalize(sec, 60, min);
  normalize(min, 60, hour);
  normalize(hour, 24, mday);
  normalize(mon, 12, year);
  if (mday < 0)
    do {
      if (mon > 0)
        mon--;
      else
        year--, mon = 11;
      mday += mdays(year, mon);
    }
    while (mday < 0);
  else
    while (mday >= (mdays = mdays(year, mon))) {
      mday -= mdays;
      if (mon < 11)
        mon++;
      else
        year++, mon = 0;
    }
  yday = month_start(year)[mon] + mday;

  t = hour * 3600 + min * 60 + sec; /* nombre de secondes depuis le debut du jour local */
  t += yday * dsecs;
  /* t est le nombre de secondes depuis le debut de l'annee locale. */
  if (year < 1970)
    for (y = 1970; y > year; y--, t -= ydays(y) * dsecs);
  else
    for (y = 1970; y < year; t += ydays(y) * dsecs, y++);
  isdst = is_dst(t + tz.offset);
  t += isdst && tz.dst ? tz.dst->offset : tz.offset; /* decalage horaire */

  return t;
}

/* Retourne l'heure systeme sous forme du nombre de secondes ecoulees depuis le
 * 1er janvier 1970 a 00h 00m 00s GMT. L'heure systeme est calculee a partir de
 * l'heure de l'horloge materielle. Il n'y a pas d'horloge logicielle. */
long get_time()
{
  struct calendar_time_struct calendar_time;

  get_calendar_time(&calendar_time);
  return make_time(&calendar_time);
}

void time_init()
{
  startup_time = get_time();
}
