/* Reference: RFC 1952 */
/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/* Table of CRCs of all 8-bit messages. */
unsigned long crc_table[256];

/* Flag: has the table been computed? Initially false. */
int crc_table_computed = 0;

/* Make the table for a fast CRC. */
void make_crc_table(void)
{
  unsigned long c;
  int n, k;

  for (n = 0; n < 256; n++) {
    c = (unsigned long)n;
    for (k = 0; k < 8; k++)
      if (c & 1)
        c = 0xedb88320L ^ (c >> 1);
      else
        c >>= 1;
    crc_table[n] = c;
  }
  crc_table_computed = 1;
}

/* Update a running crc with the bytes buf[0..len-1] and return the updated
 * crc. The crc should be initialized to zero. Pre- and post-conditioning
 * (one's complement) is performed within this function so it shouldn't be done
 * by the caller. Usage example:
 *
 *   unsigned long crc = 0L;
 *
 *   while (read_buffer(buffer, length) != EOF) {
 *     crc = update_crc(crc, buffer, length);
 *   }
 *   if (crc != original_crc) error();
 */
unsigned long update_crc(unsigned long crc, const unsigned char *buf, int len)
{
  unsigned long c = crc ^ 0xffffffffL;
  int n;

  if (!crc_table_computed)
    make_crc_table();
  for (n = 0; n < len; n++)
    c = crc_table[(c ^ buf[n]) & 0xff] ^ (c >> 8);
  return c ^ 0xffffffffL;
}

/* Return the CRC of the bytes buf[0..len-1]. */
unsigned long crc(const unsigned char *buf, int len)
{
  return update_crc(0L, buf, len);
}
