/* Decodage d'un flux suivant l'algorithme deflate */
/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/* La gestion d'erreur part du principe qu'en cas d'erreur, le programme
 * appelant detectera une somme de controle invalide. Aussi, aucune gestion
 * particuliere n'est assuree dans inflate(). Si le flux d'entree est invalide,
 * le comportement est indetermine. */

/* Reference: RFC 1951 */

#include "util.h"

void __attribute__ ((noreturn)) die(const char *msg);

/* Types de compression */
#define T_NOCMPR    0   /* no compression */
#define T_FIXED     1   /* compressed with fixed Huffman codes */
#define T_DYNAMIC   2   /* compressed with dynamic Huffman codes */
#define T_RESERVED  3   /* reserved (error) */

#define MAX_BITS    15  /* longueur maximale d'un code */

struct node_struct {
  int value;
  struct node_struct *sons;
};

static const unsigned char len_extrabits[] = {0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0};

static const unsigned short len_base[] = {3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31, 35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258};

static const unsigned char dist_extrabits[] = {0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13};

static const unsigned short dist_base[] = {1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193, 257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145, 8193, 12289, 16385, 24577};

static const unsigned char cdlen_cdlen_order[19] = {16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15};

/* Flux d'entree */

struct istream_struct {
  const void **in;
  unsigned char mask;
};

static void istream_init(struct istream_struct *istrm, const void **in)
{
  istrm->in = in;
  istrm->mask = 1;
}

/* Lit count bits dans le flux d'entree.
 * count doit etre inferieur a 32, le resultat etant renvoye dans un unsigned
 * long. Si count est superieur a 32, les bits residuels sont lus, mais perdus.
 */
static unsigned long istream_readbits(struct istream_struct *istrm, int count)
{
  unsigned long res;
  unsigned long res_mask;

  res = 0;
  res_mask = 1;
  while (count-- > 0) {
    if (*(const unsigned char *)*istrm->in & istrm->mask)
      res |= res_mask;
    istrm->mask <<= 1;
    if (!istrm->mask) {
      (*istrm->in)++;
      istrm->mask = 1;
    }
    res_mask <<= 1;
  }
  return res;
}

/* Lit count octets dans le flux d'entree.
 * count doit etre positif. Si necessaire, la fonction istream_skip() doit
 * avoir ete appelee avant pour se positionner sur le premier octet plein non
 * lu. */
static const void *istream_readbytes(struct istream_struct *istrm, int count)
{
  const void *res;

  res = *istrm->in;
  *istrm->in += count;
  return res;
}

/* Abandonne la lecture de l'octet en cours et se positionne au debut de
 * l'octet suivant. */
static void istream_skip(struct istream_struct *istrm)
{
  if (istrm->mask > 1) {
    (*istrm->in)++;
    istrm->mask = 1;
  }
}

/* Flux de sortie */

struct ostream_struct {
  void **out;
  const void *start;
};

static void ostream_init(struct ostream_struct *ostrm, void **out)
{
  ostrm->out = out;
  ostrm->start = *out;
}

static void ostream_write(struct ostream_struct *ostrm, const unsigned char *buf, unsigned int size)
{
  memcpy(*ostrm->out, buf, size);
  *ostrm->out += size;
}

static void ostream_duplicate(struct ostream_struct *ostrm, unsigned int dist, unsigned int len)
{
  if (len > dist)
    return;
  memcpy(*ostrm->out, *ostrm->out - dist, len);
  *ostrm->out += len;
}

static int ostream_len(struct ostream_struct *ostrm)
{
  return *ostrm->out - ostrm->start;
}

/* Allocateur de memoire pour les noeuds des arbres de Huffman */

#define BLOCK_SIZE ALIGN(2 * sizeof (struct node_struct))

static void *mend;
static void *mlimit;
static void *first_free_block;

void init_malloc(void *start, void *end)
{
  mend = end;
  mlimit = (void *)ALIGN((unsigned long)start);
  first_free_block = NULL;
}

static void *malloc()
{
  void *p;

  if (first_free_block) {
    p = first_free_block;
    first_free_block = *(void **)first_free_block;
  }
  else {
    if (mlimit + BLOCK_SIZE > mend)
      die("Not enough memory");
    p = mlimit;
    mlimit = mlimit + BLOCK_SIZE;
  }
  return p;
}

static void free(void *ptr)
{
  *(void **)ptr = first_free_block;
  first_free_block = ptr;
}

/* Fonctions sur les codes de Huffman */

static void set_code(struct node_struct *tree, int value, unsigned long code, int len)
{
  unsigned long mask;
  int i;

  if (len > MAX_BITS)
    return ;
  for (mask = 1 << (len - 1); mask > 0; mask >>= 1) {
    i = code & mask ? 1 : 0;
    if (!tree->sons) {
      tree->sons = malloc();
      tree->sons[0].sons = NULL;
      tree->sons[1].sons = NULL;
    }
    tree = &tree->sons[i];
  }
  tree->value = value;
}

/* Genere l'arbre d'un code a partir des longueurs des codes. Les ressources
 * allouees lors de la construction de l'arbre devront etre liberees par
 * free_tree(). */
static void generate_tree(struct node_struct *tree, const unsigned char cdlen[], unsigned int size)
{
  unsigned int max_code;
  int bl_count[MAX_BITS + 1];
  unsigned long next_code[MAX_BITS + 1];
  unsigned long code;
  int bits, len;
  unsigned int i, n;

  max_code = size - 1;
  tree->sons = NULL;

  memset(bl_count, 0, sizeof bl_count);
  for (i = 0; i < size; i++)
    if ((n = cdlen[i]) >= 1)
      bl_count[n]++;

  code = 0;
  next_code[0] = 0;
  for (bits = 1; bits <= MAX_BITS; bits++) {
    code = (code + bl_count[bits - 1]) << 1;
    next_code[bits] = code;
  }

  for (n = 0; n <= max_code; n++) {
    len = cdlen[n];
    if (len) {
      set_code(tree, n, next_code[len], len);
      next_code[len]++;
    }
  }
}

/* Libere les ressources allouees dans l'arbre d'un code. */
static void free_tree(struct node_struct *tree)
{
  if (tree->sons) {
    free_tree(&tree->sons[0]);
    free_tree(&tree->sons[1]);
    free(tree->sons);
  }
}

/* Fonctions de lecture/decodage */

static inline int read_decode(struct istream_struct *istrm, const struct node_struct *tree)
{
  while (tree->sons)
    tree = &tree->sons[istream_readbits(istrm, 1)];
  return tree->value;
}

/* Retourne -1 en cas d'erreur. */
static inline int read_len(struct istream_struct *istrm, int code)
{
  if (code < 257 || code > 285)
    return -1;
  return len_base[code - 257] + istream_readbits(istrm, len_extrabits[code - 257]);
}


/* Retourne -1 en cas d'erreur. */
static inline int read_dist(struct istream_struct *istrm, const struct node_struct *tree)
{
  int code;

  code = read_decode(istrm, tree);

  if (code > 29)
    return -1;
  return dist_base[code] + istream_readbits(istrm, dist_extrabits[code]);
}

static inline void duplicate(struct ostream_struct *ostrm, unsigned int dist, unsigned int len)
{
  while (len > 0)
    if (len <= dist) {
      ostream_duplicate(ostrm, dist, len);
      break;
    }
    else {
      ostream_duplicate(ostrm, dist, dist);
      len -= dist;
    }
}

/* Fonction inflate() */

int inflate(const void **in, void **out)
{
  struct istream_struct istrm;                  /* flux d'entree */
  struct ostream_struct ostrm;                  /* flux de sortie */
  unsigned char litlen_cdlen[288];              /* literal/length code lengths */
  unsigned char dist_cdlen[32];                 /* distance code lengths */
  unsigned char cdlen_cdlen[19];                /* code lengths code lengths */
  struct node_struct *litlen_tree;              /* literal/length code */
  static struct node_struct *fixed_litlen_tree; /* fixed literal/length code */
  struct node_struct *dist_tree;                /* distance code */
  static struct node_struct *fixed_dist_tree;   /* fixed distance code */
  struct node_struct cdlen_tree;                /* code length code */
  unsigned char bfinal;                         /* indicateur de dernier bloc */
  unsigned char btype;                          /* type de compression */
  unsigned short len;                           /* nombre d'octets de donnees dans le bloc */
  unsigned short nlen;                          /* complement a 1 de len */
  unsigned char hlit;                           /* # of Literal/Length codes - 257 */
  unsigned char hdist;                          /* # of Distance codes - 1 */
  unsigned char hclen;                          /* # of Code Length codes - 4 */
  int value;
  int length;
  int distance;
  int bndry;
  int v, pv;
  int i, j;

  litlen_tree = NULL;
  fixed_litlen_tree = NULL;
  dist_tree = NULL;
  fixed_dist_tree = NULL;

  istream_init(&istrm, in);
  ostream_init(&ostrm, out);

  do {
    /* read block header from input stream. */
    bfinal = istream_readbits(&istrm, 1);
    btype = istream_readbits(&istrm, 2);
    if (btype == T_RESERVED)
      goto end;

    if (btype == T_NOCMPR) { /* stored with no compression */
      /* skip any remaining bits in current partially processed byte */
      istream_skip(&istrm);

      /* read LEN and NLEN */
      len = *(const unsigned short *)istream_readbytes(&istrm, 2);
      nlen = *(const unsigned short *)istream_readbytes(&istrm, 2);
      if ((len + nlen + 1) & 0xffff)
        goto end;

      /* copy LEN bytes of data to output */
      ostream_write(&ostrm, istream_readbytes(&istrm, len), len);
    }
    else {
      if (btype == T_FIXED) {
        if (!fixed_litlen_tree) {
          for (i = 0; i <= 143; i++)
            litlen_cdlen[i] = 8;
          for (; i <= 255; i++)
            litlen_cdlen[i] = 9;
          for (; i <= 279; i++)
            litlen_cdlen[i] = 7;
          for (; i <= 287; i++)
            litlen_cdlen[i] = 8;
          fixed_litlen_tree = malloc();
          generate_tree(fixed_litlen_tree, litlen_cdlen, sizeof litlen_cdlen);
        }
        litlen_tree = fixed_litlen_tree;
        if (!fixed_dist_tree) {
          memset(dist_cdlen, 5, sizeof dist_cdlen);
          fixed_dist_tree = malloc();
          generate_tree(fixed_dist_tree, dist_cdlen, sizeof dist_cdlen);
        }
        dist_tree = fixed_dist_tree;
      }
      if (btype == T_DYNAMIC) { /* compressed with dynamic Huffman codes */
        /* read representation of code trees */
        hlit = istream_readbits(&istrm, 5); /* # of Literal/Length codes - 257 */
        hdist = istream_readbits(&istrm, 5); /* # of Distance codes - 1 */
        hclen = istream_readbits(&istrm, 4); /* # of Code Length codes - 4 */
        memset(cdlen_cdlen, 0, sizeof cdlen_cdlen); /* code lengths for the code length alphabet */
        for (i = 0; i < hclen + 4; i++)
          cdlen_cdlen[cdlen_cdlen_order[i]] = istream_readbits(&istrm, 3);
        generate_tree(&cdlen_tree, cdlen_cdlen, sizeof cdlen_cdlen);
        memset(litlen_cdlen, 0, sizeof litlen_cdlen); /* code lengths for the literal/length alphabet and for the distance alphabet */
        memset(dist_cdlen, 0, sizeof dist_cdlen);
        bndry = hlit + 257;
        pv = 0;
        for (i = 0; i < hlit + hdist + 258; ) {
          v = read_decode(&istrm, &cdlen_tree);
          if (v >= 0 && v <= 15) { /* represent code lengths of 0 - 15 */
            i < bndry ? (litlen_cdlen[i] = v) : (dist_cdlen[i - bndry] = v);
            i++;
            pv = v;
          }
          else if (v == 16) /* copy the previous code length 3 - 6 times */
            for (j = 3 + istream_readbits(&istrm, 2); j > 0; j--) {
              i < bndry ? (litlen_cdlen[i] = pv) : (dist_cdlen[i - bndry] = pv);
              i++;
            }
          else if (v == 17) { /* repeat a code length of 0 for 3 - 10 times */
            for (j = 3 + istream_readbits(&istrm, 3); j > 0; j--) {
              i < bndry ? (litlen_cdlen[i] = 0) : (dist_cdlen[i - bndry] = 0);
              i++;
            }
            pv = 0;
          }
          else if (v == 18) { /* repeat a code length of 0 for 11 - 138 times */
            for (j = 11 + istream_readbits(&istrm, 7); j > 0; j--) {
              i < bndry ? (litlen_cdlen[i] = 0) : (dist_cdlen[i - bndry] = 0);
              i++;
            }
            pv = 0;
          }
          else
            goto end;
        }
        litlen_tree = malloc();
        generate_tree(litlen_tree, litlen_cdlen, sizeof litlen_cdlen);
        dist_tree = malloc();
        generate_tree(dist_tree, dist_cdlen, sizeof dist_cdlen);
        free_tree(&cdlen_tree);
      }
      while (1) {
        /* decode literal/length value from input stream */
        value = read_decode(&istrm, litlen_tree);
        if (value < 0 || value > 285)
          goto end;

        if (value < 256)
          /* copy value (literal byte) to output stream */
          ostream_write(&ostrm, (unsigned char *)&value, 1);
        else {
          if (value == 256) /* end of block */
            break;
          else { /* value = 257..285 */
            length = read_len(&istrm, value);
            if (length < 0)
              goto end;

            /* decode distance from input stream */
            distance = read_dist(&istrm, dist_tree);
            if (distance < 0)
              goto end;

            /* move backwards distance bytes in the output stream, and copy
               length bytes from this position to the output stream. */
            duplicate(&ostrm, distance, length);
          }
        }
      }
      if (btype == T_DYNAMIC) {
        free_tree(litlen_tree);
        free(litlen_tree);
        free_tree(dist_tree);
        free(dist_tree);
      }
    }
  }
  while (!bfinal); /* not last block */

 end:
  istream_skip(&istrm);
  return ostream_len(&ostrm);
}
