/* Relocations. */
/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "reloc.h"

#include "resolve.h"
#include "load.h"
#include "vars.h"
#include "object_info.h"
#include "error.h"
#include "util.h"

#include <elf_ext.h>

#include <string.h>

#define object_filename(objet_info) ((object_info)->filename ? : (object_info)->name)

void lazy_bind();

static void lookup_error_print_progname()
{
  fprintf(stderr, "%s: symbol lookup error", executable_invocation_name);
}

static void relocate_entry(struct object_info_struct *object_info, Elf32_Addr offset, Elf32_Word info, Elf32_Sword addend)
{
  Elf32_Addr *rel_field;
  const Elf32_Sym *symbol;
  int flags;
  struct object_info_struct *found_object_info;
  const Elf32_Sym *found_symbol;
  Elf32_Addr sym_val;

  rel_field = object_info->base_address + offset;
  switch (ELF32_R_TYPE(info)) {
  case R_386_NONE:
    return;
  case R_386_32:
  case R_386_PC32:
  case R_386_COPY:
  case R_386_GLOB_DAT:
  case R_386_JMP_SLOT:
    break;
  case R_386_RELATIVE:
    *rel_field = (Elf32_Addr)object_info->base_address + addend;
    return;
  case R_386_GOT32:
  case R_386_PLT32:
  case R_386_GOTOFF:
  case R_386_GOTPC:
  case R_386_32PLT:
  case R_386_TLS_TPOFF:
  case R_386_TLS_IE:
  case R_386_TLS_GOTIE:
  case R_386_TLS_LE:
  case R_386_TLS_GD:
  case R_386_TLS_LDM:
  case R_386_16:
  case R_386_PC16:
  case R_386_8:
  case R_386_PC8:
  case R_386_TLS_GD_32:
  case R_386_TLS_GD_PUSH:
  case R_386_TLS_GD_CALL:
  case R_386_TLS_GD_POP:
  case R_386_TLS_LDM_32:
  case R_386_TLS_LDM_PUSH:
  case R_386_TLS_LDM_CALL:
  case R_386_TLS_LDM_POP:
  case R_386_TLS_LDO_32:
  case R_386_TLS_IE_32:
  case R_386_TLS_LE_32:
  case R_386_TLS_DTPMOD32:
  case R_386_TLS_DTPOFF32:
  case R_386_TLS_TPOFF32:
  default:
    nonfatal_error("Warning: %s: unhandled relocation type %d at offset 0x%lx", object_filename(object_info), ELF32_R_TYPE(info), offset);
    return;
  }

  symbol = &object_info->dynamic_data.symbol_table[ELF32_R_SYM(info)];
  switch (ELF32_R_TYPE(info)) {
  case R_386_COPY:
    flags = RF_NOEXEC;
    break;
  default:
    flags = 0;
  }
  if (!resolve_symbol(object_info, symbol, flags, &found_object_info, &found_symbol))
    fatal_error("%s: undefined symbol: %s", object_filename(object_info), object_info->dynamic_data.string_table + symbol->st_name);
  if (found_object_info)
    sym_val = (Elf32_Addr)(found_object_info->base_address + found_symbol->st_value);
  else
    /* Le symbole est une reference faible non resolue. */
    sym_val = 0;
  switch (ELF32_R_TYPE(info)) {
  case R_386_32:
    *rel_field = sym_val + addend;
    return;
  case R_386_PC32:
    *rel_field = sym_val + addend - (Elf32_Addr)rel_field;
    return;
  case R_386_COPY:
    if (found_symbol->st_size != symbol->st_size)
      nonfatal_error("Warning: symbol %s has different size in shared object", object_info->dynamic_data.string_table + symbol->st_name);
    memcpy(rel_field, (void *)sym_val, min(found_symbol->st_size, symbol->st_size));
    return;
  case R_386_GLOB_DAT:
  case R_386_JMP_SLOT:
    *rel_field = sym_val;
    return;
  }
}

static void relocate_entry_rel(struct object_info_struct *object_info, const Elf32_Rel *rel_entry)
{
  relocate_entry(object_info, rel_entry->r_offset, rel_entry->r_info, *(Elf32_Sword *)(object_info->base_address + rel_entry->r_offset));
}

static void relocate_entry_rela(struct object_info_struct *object_info, const Elf32_Rela *rela_entry)
{
  relocate_entry(object_info, rela_entry->r_offset, rela_entry->r_info, rela_entry->r_addend);
}

static void plt_relocate_entry(struct object_info_struct *object_info, Elf32_Addr offset, Elf32_Word info, Elf32_Sword addend)
{
  Elf32_Addr *rel_field;

  rel_field = object_info->base_address + offset;
  switch (ELF32_R_TYPE(info)) {
  case R_386_JMP_SLOT:
    *rel_field = (Elf32_Addr)object_info->base_address + addend;
    break;
  default:
    fatal_error("%s: Unexpected relocation type %d at offset 0x%lx in PLT", object_filename(object_info), ELF32_R_TYPE(info), offset);
  }
}

static void plt_relocate_entry_rel(struct object_info_struct *object_info, const Elf32_Rel *rel_entry)
{
  plt_relocate_entry(object_info, rel_entry->r_offset, rel_entry->r_info, *(Elf32_Sword *)(object_info->base_address + rel_entry->r_offset));
}

static void plt_relocate_entry_rela(struct object_info_struct *object_info, const Elf32_Rela *rela_entry)
{
  plt_relocate_entry(object_info, rela_entry->r_offset, rela_entry->r_info, rela_entry->r_addend);
}

static void relocate_table_rel(struct object_info_struct *object_info, const Elf32_Rel *relocation_table, int relocation_table_size)
{
  int i;

  for (i = 0; i < relocation_table_size; i++)
    relocate_entry_rel(object_info, &relocation_table[i]);
}

static void relocate_table_rela(struct object_info_struct *object_info, const Elf32_Rela *relocation_table, int relocation_table_size)
{
  int i;

  for (i = 0; i < relocation_table_size; i++)
    relocate_entry_rela(object_info, &relocation_table[i]);
}

static void plt_relocate_table_rel(struct object_info_struct *object_info, const Elf32_Rel *relocation_table, int relocation_table_size)
{
  int i;

  for (i = 0; i < relocation_table_size; i++)
    plt_relocate_entry_rel(object_info, &relocation_table[i]);
}

static void plt_relocate_table_rela(struct object_info_struct *object_info, const Elf32_Rela *relocation_table, int relocation_table_size)
{
  int i;

  for (i = 0; i < relocation_table_size; i++)
    plt_relocate_entry_rela(object_info, &relocation_table[i]);
}

void relocate_object(struct object_info_struct *object_info)
{
  if (object_info->dynamic_data.textrel)
    make_segments_writable(object_info);

  if (object_info->dynamic_data.global_offset_table) {
    object_info->dynamic_data.global_offset_table[1] = (Elf32_Addr)object_info;
    object_info->dynamic_data.global_offset_table[2] = (Elf32_Addr)lazy_bind;
  }

  if (object_info->dynamic_data.rel_relocation_table)
    relocate_table_rel(object_info, object_info->dynamic_data.rel_relocation_table, object_info->dynamic_data.rel_relocation_table_size);

  if (object_info->dynamic_data.rela_relocation_table)
    relocate_table_rela(object_info, object_info->dynamic_data.rela_relocation_table, object_info->dynamic_data.rela_relocation_table_size);

  if (object_info->dynamic_data.bind_now)
    switch (object_info->dynamic_data.plt_relocation_type) {
    case RT_REL:
      if (object_info->dynamic_data.plt_relocation_table.rel)
        relocate_table_rel(object_info, object_info->dynamic_data.plt_relocation_table.rel, object_info->dynamic_data.plt_relocation_table_size);
      break;
    case RT_RELA:
      if (object_info->dynamic_data.plt_relocation_table.rela)
        relocate_table_rela(object_info, object_info->dynamic_data.plt_relocation_table.rela, object_info->dynamic_data.plt_relocation_table_size);
      break;
    }
  else
    switch (object_info->dynamic_data.plt_relocation_type) {
    case RT_REL:
      if (object_info->dynamic_data.plt_relocation_table.rel)
        plt_relocate_table_rel(object_info, object_info->dynamic_data.plt_relocation_table.rel, object_info->dynamic_data.plt_relocation_table_size);
      break;
    case RT_RELA:
      if (object_info->dynamic_data.plt_relocation_table.rela)
        plt_relocate_table_rela(object_info, object_info->dynamic_data.plt_relocation_table.rela, object_info->dynamic_data.plt_relocation_table_size);
      break;
    }

  if (object_info->dynamic_data.textrel)
    restore_segments_prot(object_info);
}

Elf32_Addr relocate_plt_entry(struct object_info_struct *object_info, Elf32_Word offset)
{
  const Elf32_Rel *rel_entry;
  const Elf32_Rela *rela_entry;

  error_print_progname = lookup_error_print_progname;

  switch (object_info->dynamic_data.plt_relocation_type) {
  case RT_REL:
    if (!object_info->dynamic_data.plt_relocation_table.rel)
      fatal_error("%s: no PLT relocation table", object_filename(object_info));
    rel_entry = (const void *)object_info->dynamic_data.plt_relocation_table.rel + offset;
    relocate_entry_rel(object_info, rel_entry);
    return *(Elf32_Addr *)(object_info->base_address + rel_entry->r_offset);
  case RT_RELA:
    if (!object_info->dynamic_data.plt_relocation_table.rela)
      fatal_error("%s: no PLT relocation table", object_filename(object_info));
    rela_entry = (const void *)object_info->dynamic_data.plt_relocation_table.rela + offset;
    relocate_entry_rela(object_info, rela_entry);
    return *(Elf32_Addr *)(object_info->base_address + rela_entry->r_offset);
  case RT_NONE:
  default:
    fatal_error("%s: no PLT relocation table", object_filename(object_info));
  }
}
