/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "stdlib.h"

#include "unistd.h"

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <stddef.h>

#define BITS (sizeof (size_t) * 8)

#define MEMORY_GROW 131072 /* 128K */

/* niveaux de debogage */
enum {MC_IGNORE, MC_TELL, MC_ABORT, MC_TELL_ABORT};

/* bloc de memoire */
struct block_struct {
  size_t size; /* taille du bloc - le bit de poids faible indique si le bloc est alloue */
  union {
    struct { /* bloc libre */
      struct block_struct *previous;
      struct block_struct *next;
    };
    void *data[1]; /* bloc alloue */
  };
};

/* taille maximale des blocs et du tas */
#define MAX_HEAP_SIZE ((size_t)1 << (BITS - 1))

#define set_size(block, new_size) ((block)->size &= (size_t)1, (block)->size |= (new_size))
#define set_allocated(block)      ((block)->size |= (size_t)1)
#define clear_allocated(block)    ((block)->size &= ~(size_t)1)
#define get_size(block)           ((block)->size & ~(size_t)1)
#define is_allocated(block)       ((block)->size & (size_t)1)

#define BLOCK_HEADER_SIZE (offsetof(struct block_struct, data))

/* taille minimale d'un bloc */
static size_t min_block_size;

static void *memory_end;

/* tas */
static void *heap_start = NULL;
static size_t heap_size = 0;

/* liste des blocs libres */
static struct block_struct *first_free_table[BITS];

static inline int ispow2(size_t n)
{
  return !(n & (n - 1));
}

static inline int ulog2(size_t n)
{
  register int k;

  k = 0;
  while (n > (size_t)1 << k)
    k++;
  return k;
}

static inline size_t nextpow2(size_t n)
{
  register unsigned int k;

  n--;
  for (k = 1; k < BITS; k <<= 1)
    n = n | n >> k;
  return n + 1;
}

/* Allocation de memoire aupres du systeme */

static int init_memory(void **memory_start)
{
  void *addr;

  if ((addr = __sbrk(0)) == (void *)-1)
    return -1;
  *memory_start = memory_end = addr;
  return 0;
}

static int resize_memory(void *end)
{
  size_t incr;
  void *addr;

  if (end > memory_end) {
    incr = (end - memory_end + (MEMORY_GROW - 1)) / MEMORY_GROW * MEMORY_GROW;
    if ((addr = __sbrk(incr)) == (void *)-1)
      return -1;
    memory_end = addr + incr;
  }
  else if (end <= memory_end - MEMORY_GROW) {
    incr = (memory_end - end) / MEMORY_GROW * MEMORY_GROW;
    if ((addr = __sbrk(-incr)) == (void *)-1)
      return -1;
    memory_end = addr - incr;
  }
  return 0;
}

/* Liste des blocs vides */

static void init_first_free_table()
{
  unsigned int i;

  for (i = 0; i < BITS; i++)
    first_free_table[i] = NULL;
}

static void push_free(struct block_struct *block)
{
  int k;

  k = ulog2(block->size);
  block->previous = NULL;
  block->next = first_free_table[k];
  if (block->next)
    block->next->previous = block;
  first_free_table[k] = block;
}

static struct block_struct *pop_free(size_t size)
{
  int k;
  struct block_struct *block;

  k = ulog2(size);
  if (!(block = first_free_table[k]))
    return NULL;
  first_free_table[k] = block->next;
  if (block->next)
    block->next->previous = NULL;
  return block;
}

static void remove_free(struct block_struct *block)
{
  if (block->previous)
    block->previous->next = block->next;
  else
    first_free_table[ulog2(block->size)] = block->next;
  if (block->next)
    block->next->previous = block->previous;
}

/* Fonctions sur le tas */

static inline int is_valid(const struct block_struct *block)
{
  size_t offset, size;

  offset = (const void *)block - heap_start;
  size = get_size(block);
  return ispow2(size) && size >= min_block_size && size <= MAX_HEAP_SIZE && size <= heap_size - offset && !(offset & (size - 1));
}

static void print_heap()
{
  int addr_width, offset_width, size_width;
  const void *p;
  const struct block_struct *block;
  size_t block_size;
  int w;

  fputs("======= Heap: ==============\n", stderr);
  fprintf(stderr, "Total: %zu\n", heap_size);
  if (!heap_start)
    return;
  size_width = offset_width = addr_width = 0;
  for (p = heap_start; p != heap_start + heap_size; p += block_size) {
    block = p;
    block_size = get_size(block);
    addr_width = snprintf(NULL, 0, "%p", block);
    offset_width = snprintf(NULL, 0, "%#x", p - heap_start);
    w = snprintf(NULL, 0, "%zu", block_size);
    if (w > size_width)
      size_width = w;
    if (!is_valid(block))
      break;
  }
  for (p = heap_start; p != heap_start + heap_size; p += block_size) {
    block = p;
    block_size = get_size(block);
    fprintf(stderr, "%*p %#*x %*zu ", addr_width, block, offset_width, p - heap_start, size_width, block_size);
    if (!is_valid(block)) {
      fprintf(stderr, "[invalid block]\n");
      return;
    }
    if (is_allocated(block))
      fprintf(stderr, "%p\n", block->data);
    else
      fprintf(stderr, "%s\n", "(free)");
  }
}

static int setup_heap()
{
  min_block_size = nextpow2(sizeof (struct block_struct));
  if (init_memory(&heap_start) == -1)
    return -1;
  /* alignement de heap_start de sorte que le pointeur sur les donnees des
     blocs soit correctement aligne pour n'importe quel type d'objet (ANSI C)
  */
  heap_start += BLOCK_HEADER_SIZE;
  heap_start = (void *)(((intptr_t)heap_start | (sizeof (union {long l; double d;}) - 1)) + 1);
  heap_start -= BLOCK_HEADER_SIZE;
  heap_size = 0;
  init_first_free_table();
  return 0;
}

/* size doit etre un diviseur de heap_size. */
static int grow_heap(size_t incr)
{
  struct block_struct *block, *buddy;

  if (incr > MAX_HEAP_SIZE - heap_size) {
    errno = ENOMEM;
    return -1;
  }
  if (resize_memory(heap_start + heap_size + incr) == -1)
    return -1;
  block = heap_start + heap_size;
  block->size = incr;
  heap_size += incr;
  while (((void *)block - heap_start) & block->size && (buddy = (void *)block - block->size, !is_allocated(buddy)) && buddy->size == block->size) {
    remove_free(buddy);
    block = buddy;
    block->size <<= 1;
  }
  push_free(block);
  return 0;
}

static void shrink_heap(size_t incr)
{
  heap_size -= incr;
  resize_memory(heap_start + heap_size);
}

/* Fonctions sur les blocs */

/* Cherche un bloc libre de taille minimale size dans le tas. */
static struct block_struct *search_free(size_t size)
{
  struct block_struct *block;

  while (size <= heap_size) {
    if ((block = pop_free(size)))
      return block;
    if (size == MAX_HEAP_SIZE)
      return NULL;
    size <<= 1;
  }
  return NULL;
}

/* Extrait du tas un bloc libre de taille exactement size. */
static struct block_struct *get_free_block(size_t size)
{
  struct block_struct *block, *buddy;

  if (!(block = search_free(size)))
    return NULL;
  /* decoupage du bloc libre jusqu'a la taille adequate */
  while (block->size > size) {
    block->size >>= 1;
    buddy = (void *)block + block->size;
    buddy->size = block->size;
    push_free(buddy);
  }
  return block;
}

/* Reintegre dans le tas un bloc libre precedemment extrait. */
static void merge_block(struct block_struct *block)
{
  size_t offset, size;
  struct block_struct *buddy;

  /* fusion avec le compagnon tant que possible */
  while (1) {
    offset = (void *)block - heap_start;
    size = block->size;
    if (offset & size || offset + size < heap_size) {
      buddy = heap_start + (offset ^ size);
      if (is_allocated(buddy) || buddy->size != size)
        break;
      remove_free(buddy);
      if (buddy < block)
        block = buddy;
    }
    else { /* pas de compagnon */
      if (offset == 0) /* pas d'autre bloc dans le tas */
        break;
      heap_size += size; /* compagnon virtuel */
    }
    block->size <<= 1;
  }
  if ((void *)block + block->size < heap_start + heap_size)
    push_free(block);
  else
    shrink_heap(block->size);
}

/* Agrandit un bloc alloue. size doit etre superieur a la taille du bloc. */
static int grow_block(struct block_struct *block, size_t size)
{
  size_t offset, block_size;
  struct block_struct *buddy;

  block_size = get_size(block);
  offset = (void *)block - heap_start;
  do {
    if (offset & block_size)
      return 0;
    if (offset + block_size == heap_size)
      if (grow_heap(block_size) == -1)
        return 0;
    buddy = (void *)block + block_size;
    if (is_allocated(buddy) || buddy->size != block_size)
      return 0;
    remove_free(buddy);
    block_size <<= 1;
    set_size(block, block_size);
  }
  while (block_size < size);
  return 1;
}

/* Retricit un bloc alloue. size doit etre inferieur a la taille du bloc. */
static void shrink_block(struct block_struct *block, size_t size)
{
  size_t block_size;
  struct block_struct *buddy;

  block_size = get_size(block);
  if ((void *)block + block_size < heap_start + heap_size)
    do {
      block_size >>= 1;
      buddy = (void *)block + block_size;
      buddy->size = block_size;
      push_free(buddy);
    }
    while (block_size > size);
  else
    shrink_heap(block_size - size);
  set_size(block, size);
}

static struct block_struct *alloc_block(size_t size)
{
  struct block_struct *block;
  size_t incr;

  if (!(block = get_free_block(size))) {
    incr = min_block_size;
    while (1) {
      if (heap_size & incr || incr == size) {
        /* agrandissement du tas */
        if (grow_heap(incr) == -1)
          return NULL;
        if ((block = get_free_block(incr > size ? incr : size)))
          break;
      }
      incr <<= 1;
    }
  }
  set_allocated(block);
  return block;
}

static void free_block(struct block_struct *block)
{
  clear_allocated(block);
  merge_block(block);
}

/* Redimensionne un bloc alloue, si possible. */
static int resize_block(struct block_struct *block, size_t size)
{
  size_t block_size;

  block_size = get_size(block);

  if (size == block_size)
    return 1;
  else if (size > block_size)
    return grow_block(block, size);
  else { /* size < block_size */
    shrink_block(block, size);
    return 1;
  }
}

/* Tests d'integrite */

static void print_error(const char *func, const char *format, ...)
{
  va_list ap;

  fflush(stdout);
  fprintf(stderr, "*** MALLOC CHECK *** %s: %s(): ", program_invocation_name, func);
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  fputc('\n', stderr);
}

static int get_check_level()
{
  const char *val;

  if (!(val = getenv("MALLOC_CHECK_")) || !*val)
    return MC_TELL_ABORT; /* niveau par defaut */
  switch (val[0]) {
  case '0':
    return MC_IGNORE;
  case '1':
    return MC_TELL;
  case '2':
    return MC_ABORT;
  case '3':
    return MC_TELL_ABORT;
  default:
    return MC_IGNORE;
  }
}

static void check_ptr(const void *ptr, int check_level, const char *func)
{
  const struct block_struct *block;

  if (ptr == NULL)
    return;
  /* dans le tas ? */
  if (!heap_start || ptr < heap_start + BLOCK_HEADER_SIZE || ptr >= heap_start + heap_size)
    goto invalid_pointer;
  /* correctement aligne ? */
  block = ptr - BLOCK_HEADER_SIZE;
  if ((((const void *)block - heap_start) & (min_block_size - 1)))
    goto invalid_pointer;
  /* taille valide ? */
  if (!is_valid(block))
    goto invalid_pointer;
  /* alloue ? */
  if (!is_allocated(block))
    goto double_free;
  return;

 invalid_pointer:
  if (check_level == MC_TELL || check_level == MC_TELL_ABORT)
    print_error(func, "invalid pointer: %p", ptr);
  goto abort;

 double_free:
  if (check_level == MC_TELL || check_level == MC_TELL_ABORT)
    print_error(func, "double free or corruption: %p", ptr);
  goto abort;

 abort:
  if (check_level == MC_ABORT || check_level == MC_TELL_ABORT) {
    if (check_level == MC_TELL_ABORT)
      print_heap();
    abort();
  }
}

static void check_heap(int check_level, const char *func)
{
  const void *p;
  const struct block_struct *block;
  size_t block_size;

  if (!heap_start)
    return;
  for (p = heap_start; p != heap_start + heap_size; p += block_size) {
    block = (const struct block_struct *)p;
    block_size = get_size(block);
    if (!is_valid(block))
      goto corrupted_heap;
  }
  return;

 corrupted_heap:
  if (check_level == MC_TELL || check_level == MC_TELL_ABORT)
    print_error(func, "corrupted heap");
  if (check_level == MC_ABORT || check_level == MC_TELL_ABORT) {
    if (check_level == MC_TELL_ABORT)
      print_heap();
    abort();
  }
}

/* D'apres le standard, "If the size of the space requested is 0, the behavior
 * is implementation-defined: the value returned shall be either a null pointer
 * or a unique pointer.". Ici, nous renvoyons un pointeur unique. */
void __attribute__ ((weak, visibility ("default"))) *malloc(size_t size)
{
  int check_level;
  struct block_struct *block;

  if ((check_level = get_check_level()) != MC_IGNORE)
    check_heap(check_level, __func__);

  if (!heap_start)
    if (setup_heap() == -1)
      return NULL;

  if (size > MAX_HEAP_SIZE - BLOCK_HEADER_SIZE) {
    errno = ENOMEM;
    return NULL;
  }
  size += BLOCK_HEADER_SIZE;
  if (size < min_block_size)
    size = min_block_size;
  else
    size = nextpow2(size);
  if (!(block = alloc_block(size)))
    return NULL;
  return block->data;
}
typeof (malloc) __malloc __attribute__ ((alias ("malloc")));

void __attribute__ ((weak, visibility ("default"))) *calloc(size_t nmemb, size_t size)
{
  void *ptr;

  if (!(ptr = __malloc(nmemb * size)))
    return NULL;
  memset(ptr, 0, nmemb * size);
  return ptr;
}
typeof (calloc) __calloc __attribute__ ((alias ("calloc")));

void __attribute__ ((weak, visibility ("default"))) free(void *ptr)
{
  int check_level;

  if ((check_level = get_check_level()) != MC_IGNORE) {
    check_heap(check_level, __func__);
    check_ptr(ptr, check_level, __func__);
  }

  if (ptr == NULL)
    return;

  free_block(ptr - BLOCK_HEADER_SIZE);
}
typeof (free) __free __attribute__ ((alias ("free")));

void __attribute__ ((weak, visibility ("default"))) *realloc(void *ptr, size_t size)
{
  int check_level;
  struct block_struct *block, *new_block;
  size_t block_size;

  if (ptr == NULL)
    return __malloc(size);

  if (size == 0) {
    __free(ptr);
    return NULL;
  }

  if ((check_level = get_check_level()) != MC_IGNORE) {
    check_heap(check_level, __func__);
    check_ptr(ptr, check_level, __func__);
  }

  if (size > MAX_HEAP_SIZE - BLOCK_HEADER_SIZE) {
    errno = ENOMEM;
    return NULL;
  }
  size += BLOCK_HEADER_SIZE;
  if (size < min_block_size)
    size = min_block_size;
  else
    size = nextpow2(size);
  block = ptr - BLOCK_HEADER_SIZE;

  /* On essaie d'abord de redimensionner le bloc. */
  if (resize_block(block, size))
    return ptr;

  /* S'il n'est pas possible de redimensionner le bloc, on en alloue un nouveau
     et on copie les donnees dedans. */
  if (!(new_block = alloc_block(size)))
    return NULL;
  block_size = get_size(block);
  memcpy(new_block->data, block->data, (size < block_size ? size : block_size) -  BLOCK_HEADER_SIZE);
  free_block(block);
  return new_block->data;
}
typeof (realloc) __realloc __attribute__ ((alias ("realloc")));
