/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "builtins.h"

#include "../util.h"
#include "../safe_malloc.h"
#include "../error.h"
#include "../vars.h"

#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

enum {CM_LOGICALLY, CM_PHYSICALLY};

static const char builtin_name[] = "cd";

static char *skip_comp_fw(char *p)
{
  while (*p && *p != '/')
    p++;
  return p;
}

static char *skip_comp_bw(char *p)
{
  do
    p--;
  while (*p != '/');
  return p;
}

static char *skip_slashes(char *p)
{
  while (*p == '/')
    p++;
  return p;
}

static void delete_region(char *start, char *end)
{
  while (*end)
    *start++ = *end++;
  *start = '\0';
}

/* Le chemin d'origine doit etre un chemin absolu.
 * La forme canonique est toujours plus petite ou egale au chemin d'origine, ce
 * qui permet de travailler directement sur celui-ci sans reallouer de tampon.
 */
static void canonicalize_path(char *path)
{
  char *p; /* position courante */
  char *mark; /* fin du dernier composant a garder */
  char *comp; /* composant courant */
  int len;

  p = path;
  mark = p;
  p++;
  if (!*p)
    return;
  while (1) {
    /* barres obliques superflues */
    p = skip_slashes(p);
    if (!*p)
      break;

    /* composant */
    comp = p;
    p = skip_comp_fw(p);
    len = p - comp;
    if (len == 2 && !strncmp(comp, "..", len)) { /* dot-dot */
      if (mark != path)
        mark = skip_comp_bw(mark);
    }
    else if (len != 1 || strncmp(comp, ".", len)) { /* non-dot */
      mark++; /* barre oblique */
      if (mark != comp) {
        len = comp - mark;
        delete_region(mark, comp);
        p -= len;
      }
      mark = p;
    }
    if (!*p)
      break;

    /* barre oblique */
    p++;
    if (!*p)
      break;
  }
  if (mark == path)
    mark++;
  if (mark != p)
    delete_region(mark, p);
}

static void set_pwd(char *newpwd)
{
  if (oldpwd)
    free(oldpwd);
  oldpwd = pwd;
  pwd = newpwd;
}

static void print_usage(FILE *stream)
{
  fprintf(stream, "Usage: %s [-L|-P] [dir]\n", builtin_name);
}

static void builtin_print_usage()
{
  print_usage(stdout);
}

static void builtin_print_description()
{
  puts("Change the working directory.");
}

static int builtin_main(int argc, char *argv[])
{
  int c;
  int mode;
  const char *path;
  int print_pwd;
  char *curpath;
  size_t pwd_len;

  mode = CM_LOGICALLY;
  while ((c = getopt(argc, argv, "+LP")) != -1)
    switch (c) {
    case 'L':
      mode = CM_LOGICALLY;
      break;
    case 'P':
      mode = CM_PHYSICALLY;
      break;
    case '?':
      invalid_option_error(optopt);
      print_usage(stderr);
      return EXIT_FAILURE;
    }

  print_pwd = 0;
  if (optind < argc)
    if (!strcmp(argv[optind], "-")) {
      if (!oldpwd) {
        generic_error("OLDPWD is not set");
        return EXIT_FAILURE;
      }
      path = oldpwd;
      print_pwd = 1;
    }
    else
      path = argv[optind];
  else
    path = "/";

  if (path[0] == '/')
    curpath = safe_strdup(path);
  else {
    if (!pwd) {
      generic_error("PWD is not set");
      return EXIT_FAILURE;
    }
    pwd_len = strlen(pwd);
    curpath = safe_malloc(pwd_len + strlen(path) + 2);
    strcpy(curpath, pwd);
    curpath[pwd_len] = '/';
    strcpy(curpath + pwd_len + 1, path);
  }
  switch (mode) {
  case CM_LOGICALLY:
    canonicalize_path(curpath);
    if (chdir(curpath) == -1) {
      file_error(errno, path);
      goto error_free_curpath;
    }
    set_pwd(curpath);
    break;
  case CM_PHYSICALLY:
    if (chdir(curpath) == -1) {
      file_error(errno, path);
      goto error_free_curpath;
    }
    free(curpath);
    set_pwd(get_working_directory());
    break;
  }
  if (print_pwd)
    printf("%s\n", pwd);
  return EXIT_SUCCESS;

 error_free_curpath:
  free(curpath);
  return EXIT_FAILURE;
}

const struct builtin_struct cd_builtin = {
  .name = builtin_name,
  .print_usage = builtin_print_usage,
  .print_description = builtin_print_description,
  .func = builtin_main
};
