/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "builtins.h"

#include "util.h"
#include "../jobs.h"
#include "../error.h"

#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>

struct sig_info_struct {
  const char *name;
  int num;
};

static const char builtin_name[] = "kill";

static const struct sig_info_struct sig_table[] = {
  {"HUP", 1},
  {"INT", 2},
  {"QUIT", 3},
  {"ILL", 4},
  {"TRAP", 5},
  {"ABRT", 6},
  {"BUS", 7},
  {"FPE", 8},
  {"KILL", 9},
  {"USR1", 10},
  {"SEGV", 11},
  {"USR2", 12},
  {"PIPE", 13},
  {"ALRM", 14},
  {"TERM", 15},
  {"STKFLT", 16},
  {"CHLD", 17},
  {"CONT", 18},
  {"STOP", 19},
  {"TSTP", 20},
  {"TTIN", 21},
  {"TTOU", 22},
  {"URG", 23},
  {"XCPU", 24},
  {"XFSZ", 25},
  {"VTALRM", 26},
  {"PROF", 27},
  {"WINCH", 28},
  {"POLL", 29},
  {"PWR", 30},
  {"SYS", 31}
};

static const unsigned int sig_table_size = sizeof sig_table / sizeof *sig_table;

static void list_signal_names()
{
  const struct sig_info_struct *sig_info;

  for (sig_info = sig_table; sig_info < sig_table + sig_table_size; sig_info++)
    printf("%s%c", sig_info->name, sig_info != sig_table + (sig_table_size - 1) ? ' ' : '\n');
}

static int get_sig_from_name(const char *sig_name, int *sig)
{
  const struct sig_info_struct *sig_info;

  if (sscanf(sig_name, "%d", sig) == 1)
    return 1;
  else {
    if (!strncasecmp(sig_name, "SIG", 3))
      sig_name += 3;
    for (sig_info = sig_table; sig_info < sig_table + sig_table_size; sig_info++)
      if (!strcasecmp(sig_name, sig_info->name)) {
        *sig = sig_info->num;
        return 1;
      }
  }
  return 0;
}

static void print_usage(FILE *stream)
{
  fprintf(stream, "Usage: %s [-s sig|-sig] pid|job_id...\n", builtin_name);
  fprintf(stream, "   or: %s -l\n", builtin_name);
}

static void builtin_print_usage()
{
  print_usage(stdout);
}

static void builtin_print_description()
{
  puts("Terminate or signal processes.");
}

static int builtin_main(int argc, char *argv[])
{
  int sig;
  int optind;
  const char *arg;
  const char *sig_name;
  int status;
  struct job_struct *job;
  pid_t pid;

  sig = SIGTERM;
  optind = 1;
  while (optind < argc && argv[optind][0] == '-') {
    if (!strcmp(argv[optind], "--")) {
      optind++;
      break;
    }
    arg = argv[optind];
    if (!strcmp(arg, "-l")) {
      list_signal_names();
      return EXIT_SUCCESS;
    }
    else if (!strcmp(arg, "-s")) {
      optind++;
      if (optind == argc)
        goto error_short_option_requires_argument;
      sig_name = argv[optind];
      optind++;
      goto get_sig;
    }
    /* recherche d'une option longue */
    sig_name = arg + 1;
    optind++;
    goto get_sig;

  error_short_option_requires_argument:
    generic_error("option requires an argument -- %c", arg[1]);
    print_usage(stderr);
    return EXIT_FAILURE;

  get_sig:
    if (!get_sig_from_name(sig_name, &sig)) {
      generic_error("%s: invalid signal specification", sig_name);
      return EXIT_FAILURE;
    }
  }

  if (optind == argc) {
    generic_error("not enough arguments");
    print_usage(stderr);
    return EXIT_FAILURE;
  }
  status = EXIT_FAILURE;
  do {
    if (argv[optind][0] == '%') {
      if (!get_job_from_arg(argv[optind], &job)) {
        generic_error("%s: no such job", argv[optind]);
        continue;
      }
      if (kill_job(job, sig) == -1)
        continue;
    }
    else {
      if (!get_pid_from_arg(argv[optind], &pid)) {
        generic_error("%s: not a valid process or job ID", argv[optind]);
        continue;
      }
      if (kill(pid, sig) == -1) {
        process_error(errno, pid);
        continue;
      }
    }
    status = EXIT_SUCCESS;
  }
  while (++optind < argc);
  return status;
}

const struct builtin_struct kill_builtin = {
  .name = builtin_name,
  .print_usage = builtin_print_usage,
  .print_description = builtin_print_description,
  .func = builtin_main
};
