/* Controle des taches */
/* Copyright (C) 2008  Emmanuel Varoquaux

   This file is part of XOS.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "jobs.h"

#include "tty.h"
#include "safe_malloc.h"
#include "error.h"
#include "vars.h"

#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

/* champs d'affichage d'une tache */
#define PF_ID          0001 /* identifiant */
#define PF_CURRENT     0002 /* '+', '-' ou ' ' */
#define PF_PID         0004 /* pid du dernier processus de la conduite */
#define PF_STATE       0010 /* etat */
#define PF_TERM_STATUS 0020 /* code de retour */
#define PF_COMMAND     0040 /* commande */

#define PF_REPORT      (PF_ID | PF_CURRENT | PF_STATE | PF_COMMAND)
#define PF_FG_TERM     (PF_TERM_STATUS)
#define PF_BG_START    (PF_ID | PF_PID)
#define PF_FG          (PF_COMMAND)
#define PF_BG          (PF_ID | PF_CURRENT | PF_COMMAND)

#define JOB_TABLE_INCR 8

/* etats d'un processus */
enum {PS_RUNNING, PS_EXITED, PS_SIGNALED, PS_STOPPED, PS_UNKNOWN};

/* processus enfant du shell */
struct proc_struct {
  char *command;        /* commande executee par le processus */
  pid_t pid;            /* id du processus */
  int state;            /* etat du processus */
  union {
    int exit_status;    /* PS_EXITED */
    int term_sig;       /* PS_SIGNALED */
    int stop_sig;       /* PS_STOPPED */
  };
  struct proc_struct *next; /* prochain processus dans la conduite */
};

/* etats d'une tache */
enum {JS_RUNNING, JS_STOPPED, JS_TERMINATED, JS_UNKNOWN};

/* tache */
struct job_struct {
  int id;                      /* id de la tache */
  struct proc_struct *first_proc; /* premier processus de la conduite */
  struct proc_struct *last_proc; /* dernier processus de la conduite */
  pid_t pgid;                  /* id du groupe */
  int state;                   /* etat de la tache */
  unsigned state_changed : 1;  /* indicateur de changement d'etat */
  struct job_struct *previous; /* tache precedente dans la pile */
  struct job_struct *next;     /* tache suivante dans la pile */
};

static pid_t shell_pgid;

static struct job_struct **job_table;    /* table des taches */
static size_t job_table_size;            /* taille de la table des taches */
static struct job_struct *job_stack_top; /* sommet de la pile de taches */
static struct job_struct *least_recently_stopped_job; /* plus ancienne tache arretee */

/* Processus */

static void init_proc(struct proc_struct *proc, char *command, pid_t pid)
{
  proc->command = command;
  proc->pid = pid;
  proc->state = PS_RUNNING;
}

static void destroy_proc(struct proc_struct *proc)
{
  free(proc->command);
}

/* Positionne l'etat d'un processus a partir de l'etat retourne par wait(). */
static void set_proc_state(struct proc_struct *proc, int status)
{
  if (WIFEXITED(status)) {
    proc->state = PS_EXITED;
    proc->exit_status = WEXITSTATUS(status);
  }
  else if (WIFSIGNALED(status)) {
    proc->state = PS_SIGNALED;
    proc->term_sig = WTERMSIG(status);
  }
  else if (WIFSTOPPED(status)) {
    proc->state = PS_STOPPED;
    proc->stop_sig = WSTOPSIG(status);
  }
  else
    proc->state = PS_UNKNOWN;
}

static void update_proc_state(struct proc_struct *proc)
{
  int status;
  pid_t pid;

  if (proc->state == PS_EXITED || proc->state == PS_SIGNALED)
    return; /* processus termine */
  if ((pid = waitpid(proc->pid, &status, (interactive ? WUNTRACED : 0) | WNOHANG)) == -1) {
    sys_error(errno, "wait");
    return;
  }
  if (!pid)
    return; /* pas de changement d'etat */
  set_proc_state(proc, status);
}

static void set_proc_state_running(struct proc_struct *proc)
{
  if (proc->state == PS_EXITED || proc->state == PS_SIGNALED)
    return; /* processus termine */
  proc->state = PS_RUNNING;
}

/* Le processus doit etre dans l'etat RUNNING.
 * Apres cette fonction, il est dans l'etat EXITED, SIGNALED, STOPPED ou
 * UNKNOWN. */
static void wait_running_proc(struct proc_struct *proc)
{
  int status;

  while (waitpid(proc->pid, &status, interactive ? WUNTRACED : 0) == -1)
    if (errno != EINTR) {
      sys_error(errno, "wait");
      proc->state = PS_UNKNOWN; /* le processus ne doit plus etre dans l'etat RUNNING a la sortie de cette fonction */
      return;
    }
  set_proc_state(proc, status);
}

/* Le processus doit etre dans l'etat EXITED, SIGNALED, STOPPED ou UNKNOWN.
 * Le code d'etat est compris entre 0 et 255.
 * Le code d'etat d'un processus dont l'etat est inconnu est fixe a 128. */
static int get_proc_status(const struct proc_struct *proc)
{
  switch (proc->state) {
  case PS_EXITED:
    return proc->exit_status;
  case PS_SIGNALED:
    return 128 + proc->term_sig;
  case PS_STOPPED:
    return 128 + proc->stop_sig;
  case PS_UNKNOWN:
  default:
    return 128;
  }
}

/* Fonctions sur les processus */

static struct proc_struct *create_proc(char *command, pid_t pid)
{
  struct proc_struct *proc;

  proc = safe_malloc(sizeof (struct proc_struct));
  init_proc(proc, command, pid);
  return proc;
}

static void delete_proc(struct proc_struct *proc)
{
  destroy_proc(proc);
  free(proc);
}

/* Conduites */

static void init_pipeline(struct running_pipeline_struct *running_pipeline, int foreground)
{
  running_pipeline->foreground = foreground;
  running_pipeline->first_proc = NULL;
  running_pipeline->last_proc = NULL;
}

static void add_proc_to_pipeline(struct running_pipeline_struct *running_pipeline, struct proc_struct *proc)
{
  proc->next = NULL;
  if (!running_pipeline->first_proc) {
    running_pipeline->first_proc = proc;
    running_pipeline->last_proc = proc;
    running_pipeline->pgid = interactive ? proc->pid : shell_pgid;
  }
  else {
    running_pipeline->last_proc->next = proc;
    running_pipeline->last_proc = proc;
  }
}

static void discard_pipeline(struct running_pipeline_struct *running_pipeline)
{
  struct proc_struct *proc, *next;

  for (proc = running_pipeline->first_proc; proc; proc = next) {
    next = proc->next;
    delete_proc(proc);
  }
}

/* Table des taches */

static void init_job_table()
{
  unsigned int i;

  job_table = safe_malloc(JOB_TABLE_INCR * sizeof (struct job_struct *));
  /* job_table[0] n'est pas utilise */
  for (i = 1; i < JOB_TABLE_INCR; i++)
    job_table[i] = NULL;
  job_table_size = JOB_TABLE_INCR;
}

/* La table doit etre vide. */
static void destroy_job_table()
{
  free(job_table);
}

static void increase_job_table()
{
  unsigned int i;

  job_table = safe_realloc(job_table, (job_table_size + JOB_TABLE_INCR) * sizeof (struct job_struct *));
  for (i = job_table_size; i < job_table_size + JOB_TABLE_INCR; i++)
    job_table[i] = NULL;
  job_table_size += JOB_TABLE_INCR;
}

static void insert_job_into_table(struct job_struct *job)
{
  unsigned int i;

  i = 1;
  while (i < job_table_size) {
    if (!job_table[i])
      goto found;
    i++;
  }
  increase_job_table();
 found:
  job->id = i;
  job_table[i] = job;
}

static void remove_job_from_table(struct job_struct *job)
{
  job_table[job->id] = NULL;
}

/* Pile de taches */

static void init_job_stack()
{
  job_stack_top = NULL;
  least_recently_stopped_job = NULL;
}

static void push_job(struct job_struct *job)
{
  job->previous = job_stack_top;
  job->next = NULL;
  if (job_stack_top)
    job_stack_top->next = job;
  job_stack_top = job;
}

static void insert_job_into_stack(struct job_struct *job, struct job_struct *next)
{
  job->previous = next->previous;
  job->next = next;
  if (next->previous)
    next->previous->next = job;
  next->previous = job;
}

static void remove_job_from_stack(struct job_struct *job)
{
  if (job->previous)
    job->previous->next = job->next;
  if (job->next)
    job->next->previous = job->previous;
  else
    job_stack_top = job->previous;
}

static void push_stopped_job(struct job_struct *job)
{
  push_job(job);
  if (!least_recently_stopped_job)
    least_recently_stopped_job = job;
}

static void push_running_job(struct job_struct *job)
{
  if (least_recently_stopped_job)
    insert_job_into_stack(job, least_recently_stopped_job);
  else
    push_job(job);
}

static void remove_job(struct job_struct *job)
{
  if (job == least_recently_stopped_job)
    least_recently_stopped_job = job->next;
  remove_job_from_stack(job);
}

/* Taches */

static void init_job(struct job_struct *job, struct running_pipeline_struct *running_pipeline)
{
  insert_job_into_table(job);
  job->first_proc = running_pipeline->first_proc;
  job->last_proc = running_pipeline->last_proc;
  job->pgid = running_pipeline->pgid;
  job->state = JS_RUNNING;
  job->state_changed = 0;
  push_running_job(job);
}

static void destroy_job(struct job_struct *job)
{
  struct proc_struct *proc, *next;

  remove_job(job);
  for (proc = job->first_proc; proc; proc = next) {
    next = proc->next;
    delete_proc(proc);
  }
  remove_job_from_table(job);
}

static int get_resulting_job_state(int job_state, int proc_state)
{
  switch (job_state) {
  case JS_RUNNING:
    return JS_RUNNING;
  case JS_STOPPED:
    if (proc_state == PS_RUNNING)
      return JS_RUNNING;
    return JS_STOPPED;
  case JS_TERMINATED:
    switch (proc_state) {
    case PS_RUNNING:
      return JS_RUNNING;
    case PS_STOPPED:
      return JS_STOPPED;
    }
    return JS_TERMINATED;
  case JS_UNKNOWN:
  default:
    switch (proc_state) {
    case PS_RUNNING:
      return JS_RUNNING;
    case PS_STOPPED:
      return JS_STOPPED;
    case PS_EXITED:
    case PS_SIGNALED:
      return JS_TERMINATED;
    }
    return JS_UNKNOWN;
  }
}

static void update_job_state(struct job_struct *job)
{
  int old_state;
  struct proc_struct *proc;

  old_state = job->state;
  job->state = JS_UNKNOWN;
  for (proc = job->first_proc; proc; proc = proc->next) {
    update_proc_state(proc);
    job->state = get_resulting_job_state(job->state, proc->state);
  }
  /* La tache ne peut pas etre devenue RUNNING si elle ne l'etait pas deja. */
  if (job->state != old_state) {
    if (old_state == JS_STOPPED) {
      remove_job(job);
      push_running_job(job);
    }
    else if (job->state == JS_STOPPED) {
      remove_job(job);
      push_stopped_job(job);
    }
    job->state_changed = 1;
  }
}

/* La tache ne doit pas etre dans l'etat TERMINATED. */
static void set_job_state_running(struct job_struct *job)
{
  int old_state;
  struct proc_struct *proc;

  old_state = job->state;
  for (proc = job->first_proc; proc; proc = proc->next)
    set_proc_state_running(proc);
  job->state = JS_RUNNING; /* au moins un processus n'est pas dans l'etat TERMINATED */
  if (old_state == JS_STOPPED) {
    remove_job(job);
    push_running_job(job);
  }
  job->state_changed = 0; /* pas de report */
}

/* La tache doit etre dans l'etat STOPPED, TERMINATED ou UNKNOWN. */
static int get_job_status(const struct job_struct *job)
{
  return get_proc_status(job->last_proc);
}

/* La tache doit etre dans l'etat RUNNING.
 * Apres cette fonction, elle est dans l'etat STOPPED, TERMINATED ou UNKNOWN.
 */
static void wait_running_job(struct job_struct *job, int *status)
{
  struct proc_struct *proc;

  job->state = JS_UNKNOWN;
  for (proc = job->first_proc; proc; proc = proc->next) {
    if (proc->state == PS_RUNNING)
      wait_running_proc(proc);
    job->state = get_resulting_job_state(job->state, proc->state);
  }
  if (job->state == JS_STOPPED) {
    remove_job(job);
    push_stopped_job(job);
  }
  job->state_changed = 1; /* la tache ne peut plus etre dans l'etat RUNNING */
  *status = get_job_status(job);
}

/* Affichage */

static inline int pcount(int retval)
{
  return retval >= 0 ? retval : 0;
}

static void fill(int n, FILE *stream)
{
  fprintf(stream, "%*s", n, "");
}

static int print_job_id(const struct job_struct *job, FILE *stream)
{
  return pcount(fprintf(stream, "[%d]", job->id));
}

static int print_current_char(const struct job_struct *job, FILE *stream)
{
  char c;

  if (job == get_current_job())
    c = '+';
  else if (job_stack_top && job == get_previous_job())
    c = '-';
  else
    c = ' ';
  return pcount(fprintf(stream, "%c", c));
}

static int print_exit_status(int exit_status, FILE *stream)
{
  return pcount(exit_status ? fprintf(stream, "Exit %d", exit_status) : fprintf(stream, "Done"));
}

static int print_proc_state(const struct proc_struct *proc, FILE *stream)
{
  switch (proc->state) {
  case PS_RUNNING:
    return pcount(fprintf(stream, "Running"));
  case PS_EXITED:
    return print_exit_status(proc->exit_status, stream);
  case PS_SIGNALED:
    return pcount(fprintf(stream, "%s", strsignal(proc->term_sig)));
  case PS_STOPPED:
    return pcount(fprintf(stream, "%s", strsignal(proc->stop_sig)));
  case PS_UNKNOWN:
  default:
    return pcount(fprintf(stream, "Unknown status"));
  }
}

static int print_job_state(const struct job_struct *job, FILE *stream)
{
  switch (job->state) {
  case JS_RUNNING:
    return pcount(fprintf(stream, "Running"));
  case JS_STOPPED:
    return pcount(fprintf(stream, "Stopped"));
  case JS_TERMINATED:
    return print_proc_state(job->last_proc, stream);
  case JS_UNKNOWN:
  default:
    return pcount(fprintf(stream, "Unknown status"));
  }
}

static void print_pipeline_command(const struct job_struct *job, FILE *stream)
{
  struct proc_struct *proc;

  fprintf(stream, "%s", job->first_proc->command);
  for (proc = job->first_proc->next; proc; proc = proc->next)
    fprintf(stream, " | %s", proc->command);
  if (job->state == JS_RUNNING)
    fprintf(stream, " &");
}

static void pad(int i, int n, FILE *stream)
{
  if (i < n)
    fill(n - i, stream);
  else
    fprintf(stream, " ");
}

static void print_job(const struct job_struct *job, int format, FILE *stream)
{
  int n;

  if (format & PF_ID) {
    print_job_id(job, stream);
    fprintf(stream, " ");
    if (format & PF_CURRENT) {
      print_current_char(job, stream);
      fprintf(stream, " ");
    }
  }
  if (format & PF_PID) {
    fprintf(stream, "%d", job->last_proc->pid);
    if (format & (PF_STATE | PF_COMMAND))
      fprintf(stream, " ");
  }
  if (format & PF_STATE) {
    n = print_job_state(job, stream);
    if (format & PF_COMMAND) {
      pad(n, 20, stream);
      fprintf(stream, " ");
    }
  }
  else if (format & PF_TERM_STATUS) {
    n = print_proc_state(job->last_proc, stream);
    if (format & PF_COMMAND) {
      pad(n, 20, stream);
      fprintf(stream, " ");
    }
  }
  if (format & PF_COMMAND)
    print_pipeline_command(job, stream);
  fprintf(stream, "\n");
  fflush(stream);
}

static void print_job_long(const struct job_struct *job, FILE *stream)
{
  int n, p;
  const struct proc_struct *proc;

  n = print_job_id(job, stream);
  n += pcount(fprintf(stream, " "));
  n += print_current_char(job, stream);
  fprintf(stream, " %d ", job->pgid);
  p = print_proc_state(job->first_proc, stream);
  pad(p, 20, stream);
  fprintf(stream, "   %s", job->first_proc->command);
  for (proc = job->first_proc->next; proc; proc = proc->next) {
    fprintf(stream, "\n");
    fill(n, stream);
    fprintf(stream, " %d ", proc->pid);
    p = print_proc_state(proc, stream);
    pad(p, 20, stream);
    fprintf(stream, " | %s", proc->command);
  }
  if (job->state == JS_RUNNING)
    fprintf(stream, " &");
  fprintf(stream, "\n");
  fflush(stream);
}

static void print_job_leader(const struct job_struct *job, FILE *stream)
{
  fprintf(stream, "%d\n", job->first_proc->pid);
  fflush(stream);
}

static void print_pseudo_job(int exit_status, FILE *stream)
{
  print_exit_status(exit_status, stream);
  fprintf(stream, "\n");
  fflush(stream);
}

/* Fonctions sur les taches */

static void update_job_states()
{
  unsigned int i;

  for (i = 1; i < job_table_size; i++)
    if (job_table[i])
      update_job_state(job_table[i]);
}

static struct job_struct *create_job(struct running_pipeline_struct *running_pipeline)
{
  struct job_struct *job;

  job = safe_malloc(sizeof (struct job_struct));
  init_job(job, running_pipeline);
  return job;
}

static void delete_job(struct job_struct *job)
{
  destroy_job(job);
  free(job);
}

static void delete_jobs()
{
  unsigned int i;

  for (i = 1; i < job_table_size; i++)
    if (job_table[i])
      delete_job(job_table[i]);
}

static void report_job(struct job_struct *job, int format, FILE *stream)
{
  switch (format) {
  case RF_NORMAL:
    print_job(job, PF_REPORT, stream);
    break;
  case RF_LONG:
    print_job_long(job, stream);
    break;
    case RF_ONLY_LEADER:
      print_job_leader(job, stream);
      break;
  }
  if (job->state == JS_TERMINATED)
    delete_job(job);
  else
    job->state_changed = 0;
}

static void report_jobs(int format, FILE *stream)
{
  unsigned int i;

  for (i = 1; i < job_table_size; i++)
    if (job_table[i])
      report_job(job_table[i], format, stream);
}

static void report_unreported_jobs(FILE *stream)
{
  unsigned int i;

  for (i = 1; i < job_table_size; i++)
    if (job_table[i])
      if (job_table[i]->state_changed)
        report_job(job_table[i], RF_NORMAL, stream);
}

static void report_foreground_job(struct job_struct *job, FILE *stream)
{
  if (job->state == JS_TERMINATED)
    switch (job->last_proc->state) {
    case PS_EXITED:
      if (job->last_proc->exit_status)
        print_job(job, PF_FG_TERM, stream);
      break;
    case PS_SIGNALED:
      if (job->last_proc->term_sig == SIGINT) {
        putchar('\n');
        fflush(stdout);
      }
      else
        print_job(job, PF_FG_TERM, stream);
      break;
    }
  if (job->state == JS_TERMINATED)
    delete_job(job);
}

static void report_pseudo_job(int exit_status, FILE *stream)
{
  if (exit_status)
    print_pseudo_job(exit_status, stream);
}

static void cleanup_dead_jobs()
{
  unsigned int i;

  for (i = 1; i < job_table_size; i++)
    if (job_table[i])
      if (job_table[i]->state == JS_TERMINATED)
        delete_job(job_table[i]);
}

/* Signaux */

static void ignore_signals()
{
  signal(SIGQUIT, SIG_IGN);

  if (interactive) {
    signal(SIGINT, SIG_IGN);
    signal(SIGTERM, SIG_IGN);

    /* controle des taches */
    signal(SIGTSTP, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
  }
}

static void default_signals()
{
  signal(SIGQUIT, SIG_DFL);

  if (interactive) {
    signal(SIGTERM, SIG_DFL);
    signal(SIGINT, SIG_DFL);

    /* controle des taches */
    signal(SIGTSTP, SIG_DFL);
    signal(SIGTTIN, SIG_DFL);
    signal(SIGTTOU, SIG_DFL);
  }
}

/* Terminal de controle */

static void setup_tty()
{
  set_tty_pgid(shell_pgid);
  save_tty_attributes();
  disable_tty_signals();
}

static void restore_tty()
{
  restore_tty_attributes();
}

static void acquire_tty()
{
  set_tty_pgid(shell_pgid);
  disable_tty_signals();
}

static void acquire_tty_child()
{
  set_tty_pgid(getpgrp());
  enable_tty_signals();
}

static void give_tty_to_child(pid_t pgid)
{
  enable_tty_signals();
  set_tty_pgid(pgid);
}

/* Fonctions utiles */

static void terminate_running_pipeline(const struct running_pipeline_struct *running_pipeline)
{
  killpg(running_pipeline->pgid, SIGTERM);
  killpg(running_pipeline->pgid, SIGCONT);
}

static void restart_job(struct job_struct *job)
{
  killpg(job->pgid, SIGCONT);
  set_job_state_running(job);
}

static void wait_foreground_job(struct job_struct *job, int *status)
{
  wait_running_job(job, status);
  if (interactive) {
    acquire_tty();
    report_foreground_job(job, stderr);
  }
  else
    delete_job(job);
}

static void notify()
{
  update_job_states();
  if (interactive)
    report_unreported_jobs(stderr);
  else
    cleanup_dead_jobs();
}

static void terminate_job(const struct job_struct *job)
{
  killpg(job->pgid, SIGTERM);
  killpg(job->pgid, SIGCONT);
}

static void terminate_stopped_jobs()
{
  unsigned int i;

  for (i = 1; i < job_table_size; i++)
    if (job_table[i])
      if (job_table[i]->state == JS_STOPPED)
        terminate_job(job_table[i]);
}

/* Fonctions exportees */

/* Le controle des taches n'est actif qu'en mode interactif.
 * En mode non interactif, les taches ne sont pas placees dans des groupes
 * distincts, et leur comportement vis-a-vis de tout terminal est ignore.
 * L'interruption, la suspension et la reprise de taches au clavier, ainsi que
 * les notifications de changement d'etat des taches, sont desactivees. */

void init_job_control()
{
  ignore_signals();

  if (interactive) {
    tty = fileno(stderr);
    wait_foreground();
    if (setpgid(0, 0) == -1)
      sys_error(errno, "setpgid");

    shell_pgid = getpgrp();
    setup_tty();
  }
  else
    shell_pgid = getpgrp();

  init_job_table();
  init_job_stack();
}

void end_job_control()
{
  terminate_stopped_jobs();

  delete_jobs();
  destroy_job_table();

  if (interactive)
    restore_tty();
}

void init_running_pipeline(struct running_pipeline_struct *running_pipeline, int foreground)
{
  init_pipeline(running_pipeline, foreground);
}

/* La propriete de la commande est transferee a la tache. */
pid_t make_child(struct running_pipeline_struct *running_pipeline, char *command)
{
  pid_t pid, pgid;
  struct proc_struct *proc;

  /* creation du fils */
  if ((pid = fork()) == -1) {
    sys_error(errno, "failed to fork");
    free(command);
    return -1;
  }

  /* Le pere et le fils vont tous deux tenter de placer le fils dans un nouveau
     groupe. Au moins l'un des deux y parviendra.
     Le setpgid() du pere echoue si :
     - [EACCES] le fils a deja execute exec().
     Le pere est oblige de faire setpgid() pour :
     - etre sur que le leader du groupe existe encore au moment du setpgid()
       (puisque wait() n'a pas ete appele) ;
     - etre sur qu'a partir de ce point le fils appartient bien au bon groupe.
     Le setpgid() du fils echoue si :
     - [EPERM] le fils n'est pas leader du groupe et le groupe n'existe plus.
     Le fils est oblige de faire setpgid() pour :
     - etre sur que setpgid() sera fait avant exec(). */
  /* Le premier processus cree devient leader du groupe. */
  pgid = running_pipeline->first_proc ? running_pipeline->pgid : pid;
  if (!pid) { /* fils */
    default_signals();

    /* integration du groupe de la conduite */
    if (interactive) {
      if (setpgid(pid, pgid) == -1 && errno != EPERM)
        sys_error(errno, "setpgid");
      if (running_pipeline->foreground)
        /* Chaque processus du groupe place le groupe en avant-plan du
           terminal. Chacun d'eux est donc assure d'etre dans le groupe
           d'avant-plan avant d'executer exec(). */
        acquire_tty_child();
    }

    /* liberation des ressources heritees du pere */
    free(command);
    discard_pipeline(running_pipeline);
    delete_jobs();
    shell_pgid = pgid;
  }
  else { /* pere */
    /* placement du fils dans le groupe de la conduite */
    if (interactive)
      if (setpgid(pid, pgid) == -1 && errno != EACCES)
        sys_error(errno, "setpgid");

    /* "ajout" d'un processus dans la conduite */
    proc = create_proc(command, pid);
    add_proc_to_pipeline(running_pipeline, proc);
  }
  return pid;
}

/* Cree une tache a partir d'une conduite, l'execute en avant-plan ou en
 * arriere-plan, et retourne son code d'etat.
 * La propriete des processus de la conduite est transferee a la tache. */
int start_job(struct running_pipeline_struct *running_pipeline)
{
  struct job_struct *job;
  int status;

  job = create_job(running_pipeline);
  if (running_pipeline->foreground) {
    wait_foreground_job(job, &status);
    notify();
    return status;
  }
  else {
    if (interactive)
      print_job(job, PF_BG_START, stderr);
    return 0;
  }
}

/* Detruit une conduite en tuant ses processus. Cette fonction est a appeler
 * lorsqu'une erreur s'est produite avant qu'une tache n'ait pu etre creee. */
void kill_running_pipeline(struct running_pipeline_struct *running_pipeline)
{
  terminate_running_pipeline(running_pipeline);
  discard_pipeline(running_pipeline);
}

/* Retourne un code d'etat comme si on avait execute une tache en avant-plan.
 */
int exit_pseudo_job(int exit_status)
{
  if (interactive)
    report_pseudo_job(exit_status, stderr);
  return exit_status;
}

struct job_struct *get_job(int n)
{
  if (n < 1 || (size_t)n >= job_table_size)
    return NULL;
  return job_table[n];
}

struct job_struct *get_current_job()
{
  return job_stack_top;
}

struct job_struct *get_previous_job()
{
  return job_stack_top ? job_stack_top->previous : NULL;
}

struct job_struct *get_job_by_pid(pid_t pid)
{
  unsigned int i;
  struct proc_struct *proc;

  for (i = 1; i < job_table_size; i++)
    if (job_table[i])
      for (proc = job_table[i]->first_proc; proc; proc = proc->next)
        if (proc->pid == pid)
          return job_table[i];
  return NULL;
}

void display_job(struct job_struct *job, int format)
{
  update_job_state(job);
  report_job(job, format, stdout);
}

void display_jobs(int format)
{
  update_job_states();
  report_jobs(format, stdout);
}

void display_unreported_jobs()
{
  notify();
}

int run_job_in_foreground(struct job_struct *job, int *status)
{
  if (!interactive) {
    generic_error("no job control");
    return -1;
  }

  update_job_state(job);
  if (job->state == JS_TERMINATED) {
    generic_error("job has terminated");
    return -1;
  }
  print_job(job, PF_FG, stderr);
  give_tty_to_child(job->pgid);
  if (job->state != JS_RUNNING)
    restart_job(job);
  wait_foreground_job(job, status);
  return 0;
}

int run_job_in_background(struct job_struct *job)
{
  if (!interactive) {
    generic_error("no job control");
    return -1;
  }

  update_job_state(job);
  if (job->state == JS_RUNNING) {
    generic_error("job already in background");
    return 0;
  }
  if (job->state == JS_TERMINATED) {
    generic_error("job has terminated");
    return -1;
  }
  print_job(job, PF_BG, stderr);
  restart_job(job);
  return 0;
}

int kill_job(struct job_struct *job, int sig)
{
  update_job_state(job);
  if (killpg(job->pgid, sig) == -1) {
    process_error(errno, job->pgid);
    return -1;
  }
  if (job->state == JS_STOPPED && sig == SIGCONT)
    set_job_state_running(job);
  else if (job->state == JS_STOPPED && (sig == SIGTERM || sig == SIGHUP))
    killpg(job->pgid, SIGCONT);
  return 0;
}

void wait_job(struct job_struct *job, int *status)
{
  update_job_state(job);
  if (job->state == JS_RUNNING)
    wait_running_job(job, status);
  else
    *status = get_job_status(job);
}

void wait_jobs()
{
  unsigned int i;
  int status;

  update_job_states();
  for (i = 1; i < job_table_size; i++)
    if (job_table[i])
      if (job_table[i]->state == JS_RUNNING)
        wait_running_job(job_table[i], &status);
}
